# mypy: disable-error-code="attr-defined"
import warnings
from functools import wraps
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, cast

import boto3

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.services import ClassicELB


# ----------
# Decorators
# ----------


def add_attributes_for_get(
    func: Callable[..., "ClassicELB"],
) -> Callable[..., "ClassicELB"]:
    """
    Wraps :py:meth:`botocraft.services.elb.ClassicELBManager.get` to add the
    load balancer attributes to the :py:class:`ClassicELB` object.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "ClassicELB":
        elb = func(self, *args, **kwargs)
        if elb:
            attrs = self.describe_attributes(LoadBalancerName=elb.LoadBalancerName)
            if attrs.CrossZoneLoadBalancing:
                elb.CrossZoneLoadBalancing = attrs.CrossZoneLoadBalancing.Enabled
            if attrs.ConnectionDraining:
                elb.ConnectionDraining = attrs.ConnectionDraining
            if attrs.AccessLog:
                elb.AccessLog = attrs.AccessLog
            if attrs.ConnectionSettings:
                elb.ConnectionSettings = attrs.ConnectionSettings
            if attrs.AdditionalAttributes:
                elb.AdditionalAttributes = {
                    attr.Key: attr.Value for attr in attrs.AdditionalAttributes
                }
        return elb

    return wrapper


def add_attributes_for_list(
    func: Callable[..., "PrimaryBoto3ModelQuerySet"],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.elb.ClassicELBManager.list` to add the
    load balancer attributes to the :py:class:`ClassicELB` objects.

    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        qs = func(self, *args, **kwargs)
        if qs.results:
            for elb in qs.results:
                attrs = self.describe_attributes(LoadBalancerName=elb.LoadBalancerName)
                if attrs.CrossZoneLoadBalancing:
                    elb.CrossZoneLoadBalancing = attrs.CrossZoneLoadBalancing.Enabled
                if attrs.ConnectionDraining:
                    elb.ConnectionDraining = attrs.ConnectionDraining
                if attrs.AccessLog:
                    elb.AccessLog = attrs.AccessLog
                if attrs.ConnectionSettings:
                    elb.ConnectionSettings = attrs.ConnectionSettings
                if attrs.AdditionalAttributes:
                    elb.AdditionalAttributes = {
                        attr.Key: attr.Value for attr in attrs.AdditionalAttributes
                    }
        return qs

    return wrapper


def add_tags_for_get(
    func: Callable[..., Optional["ClassicELB"]],
) -> Callable[..., Optional["ClassicELB"]]:
    """
    Wraps :py:meth:`botocraft.services.ecs.ServiceManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Service` objects instead of only a list of
    ARNs.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Optional["ClassicELB"]:
        elb = func(self, *args, **kwargs)
        if elb:
            tags = elb.objects.using(self.session).describe_tags(
                LoadBalancerNames=[elb.LoadBalancerName]
            )
            tags = tags[0].Tags
            elb.Tags = {tag.Key: tag.Value for tag in tags}
        return elb

    return wrapper


def add_tags_for_list(
    func: Callable[..., "PrimaryBoto3ModelQuerySet"],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.ecs.ServiceManager.list` to return a list of
    :py:class:`botocraft.services.ecs.Service` objects instead of only a list of
    ARNs.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        # Avoid circular import
        from botocraft.services import ClassicELB

        elbs = func(self, *args, **kwargs)
        if elbs:
            elb_names = [elb.LoadBalancerName for elb in elbs]
            tags = ClassicELB.objects.using(self.session).describe_tags(
                LoadBalancerNames=[elb_names]
            )
            for elb in elbs:
                for tag in tags:
                    if tag["LoadBalancerName"] == elb.LoadBalancerName:
                        elb.Tags = {tag.Key: tag.Value for tag in tag.Tags}
        return elbs

    return wrapper


# ----------
# Decorators
# ----------


class ClassicELBManagerMixin:
    """
    A mixin is used on :py:class:`botocraft.services.elb.ClassicELBManager` to
    implement the "create" method.   Normally we would allow the "create"
    to be automatically generated by the botocraft library, but the Classic ELB
    is fragmented across multiple API calls, so we have to do it here.

    Important:
        Note that there is no "update" method for Classic ELBs.  If you want to
        update a Classic ELB, use the methods on the :py:class:`ClassicELB` object
        itself.

    """

    session: boto3.session.Session

    def create(
        self,
        elb: "ClassicELB",
    ) -> "ClassicELB":
        """
        Create a new :py:class:`ClassicELB` object.  The reason we have to do this
        in a mixin instead of having this as a usual ``.create()`` method automatically
        generated is because Classic ELBs are fragmented across multiple API calls.

        We want to pretend that the Classic ELB is a single object, so we have to
        deal with the fragmentation here.

        Args:
            elb: The :py:class:`ClassicELB` object to create

        Returns:
            The newly created :py:class:`ClassicELB` object

        """
        tags: List[Dict[str, str]] | None = None
        if elb.Tags:
            tags = [{"Key": key, "Value": value} for key, value in elb.Tags.items()]
        # Since we use .create() directly on the manager, any .using(session) should
        # have taken place already, so we can just use self.client
        self.client.create_load_balancer(
            LoadBalancerName=elb.LoadBalancerName,
            Listeners=self.serialize(
                [listener.Listeners for listener in elb.Listeners]
            ),
            AvailabilityZones=elb.AvailabilityZones,
            Subnets=elb.Subnets,
            SecurityGroups=elb.SecurityGroups,
            Scheme=elb.Scheme,
            Tags=tags,
        )

        if elb.HealthCheck:
            self.configure_health_check(
                LoadBalancerName=elb.LoadBalancerName,
                HealthCheck=self.serialize(elb.HealthCheck),
            )

        if elb.Policies:
            if elb.Policies:
                if elb.Policies.AppCookieStickinessPolicies:
                    self.client.create_app_cookie_stickiness_policy(
                        LoadBalancerName=elb.LoadBalancerName,
                        PolicyName=elb.Policies.AppCookieStickinessPolicies[
                            0
                        ].PolicyName,
                        CookieName=elb.Policies.AppCookieStickinessPolicies[
                            0
                        ].CookieName,
                    )
                if elb.Policies.LBCookieStickinessPolicies:
                    self.client.create_lb_cookie_stickiness_policy(
                        LoadBalancerName=elb.LoadBalancerName,
                        PolicyName=elb.Policies.LBCookieStickinessPolicies[
                            0
                        ].PolicyName,
                        CookieExpirationPeriod=elb.Policies.LBCookieStickinessPolicies[
                            0
                        ].CookieExpirationPeriod,
                    )
                if elb.Policies.OtherPolicies:
                    names = ", ".join(elb.Policies.OtherPolicies)
                    # Warn that these policies must be created via
                    # ClassicELB.create_policy
                    warnings.warn(
                        f"ClassicELB(LoadBalancerName='{elb.LoadBalancerName}'): "
                        "These polices must be created via ClassicELB.create_policy: "
                        f"{names}",
                        UserWarning,
                        stacklevel=2,
                    )

        # Attributes
        self.update_attributes(elb)
        return self.get(elb.LoadBalancerName)

    def update_attributes(self, elb: "ClassicELB") -> None:
        """
        Update the attributes on an existing :py:class:`ClassicELB` object.

        Args:
            elb: The existing :py:class:`ClassicELB`

        """
        from botocraft.services import (
            AdditionalAttribute,
            ClassicELBCrossZoneLoadBalancing,
            ClassicELBLoadBalancerAttributes,
        )

        additional_attributes = None
        if elb.AdditionalAttributes:
            additional_attributes = [
                AdditionalAttribute(Key=key, Value=value)
                for key, value in elb.AdditionalAttributes.items()
            ]
        attrs = ClassicELBLoadBalancerAttributes(
            CrossZoneLoadBalancing=ClassicELBCrossZoneLoadBalancing(
                Enabled=elb.CrossZoneLoadBalancing
            ),
            ConnectionDraining=elb.ConnectionDraining,
            AccessLog=elb.AccessLog,
            ConnectionSettings=elb.ConnectionSettings,
            AdditionalAttributes=additional_attributes,
        )
        self.modify_attributes(
            LoadBalancerName=elb.LoadBalancerName,
            LoadBalancerAttributes=self.serialize(attrs),
        )

    def update_tags(self, old_elb: "ClassicELB", new_elb: "ClassicELB") -> None:
        """
        Update the tags on an existing :py:class:`ClassicELB` object.

        Args:
            old_elb: The existing :py:class:`ClassicELB
            new_elb: The new :py:class:`ClassicELB` object

        """
        manager = self.using(new_elb.session)
        if old_elb.Tags != new_elb.Tags:
            if not old_elb.Tags and new_elb.Tags:
                # Just add all the tags
                manager.add_tags(
                    LoadBalancerNames=[new_elb.LoadBalancerName],
                    Tags=[
                        {"Key": key, "Value": new_elb.Tags[key]} for key in new_elb.Tags
                    ],
                )
            elif old_elb.Tags and not new_elb.Tags:
                # Just remove all the tags
                manager.remove_tags(
                    LoadBalancerNames=[new_elb.LoadBalancerName],
                    Tags=[{"Key": key} for key in old_elb.Tags],
                )
            else:
                old_tags = cast("Dict[str, str]", old_elb.Tags)
                new_tags = cast("Dict[str, str]", new_elb.Tags)
                # First get the tags that need to be deleted by finding keys that
                # are in old_elb.Tags but not in new_elb.Tags
                keys_to_delete = set(old_tags.keys()) - set(new_tags.keys())
                if keys_to_delete:
                    manager.remove_tags(
                        LoadBalancerNames=[new_elb.LoadBalancerName],
                        Tags=[{"Key": key} for key in keys_to_delete],
                    )
                # Now get the tags that need to be added by finding keys that
                # are in new_elb.Tags but not in old_elb.Tags
                keys_to_add = set(new_tags.keys()) - set(old_tags.keys())
                if keys_to_add:
                    manager.add_tags(
                        LoadBalancerNames=[new_elb.LoadBalancerName],
                        Tags=[
                            {"Key": key, "Value": new_tags[key]} for key in keys_to_add
                        ],
                    )
                # Now find the tags that need to be updated by finding keys that
                # are in both new_elb.Tags and old_elb.Tags but have different values
                keys_to_update = set(new_tags.keys()) & set(old_tags.keys())
                keys_to_update = {
                    key for key in keys_to_update if new_tags[key] != old_tags.Tags[key]
                }
                if keys_to_update:
                    # There is no update_tags, so we have to remove and re-add
                    manager.remove_tags(
                        LoadBalancerNames=[new_elb.LoadBalancerName],
                        Tags=[{"Key": key} for key in keys_to_update],
                    )
                    manager.add_tags(
                        LoadBalancerNames=[new_elb.LoadBalancerName],
                        Tags=[
                            {"Key": key, "Value": new_tags[key]}
                            for key in keys_to_update
                        ],
                    )

    def update_subnets(self, old_elb: "ClassicELB", new_elb: "ClassicELB") -> None:
        """
        Update the subnets on an existing :py:class:`ClassicELB` object.

        Args:
            old_elb: The existing :py:class:`ClassicELB
            new_elb: The new :py:class:`ClassicELB` object

        """
        manager = self.using(old_elb.session)
        old_subnets = set()
        if old_elb.Subnets:
            old_subnets = set(old_elb.Subnets)
        new_subnets = set()
        if new_elb.Subnets:
            new_subnets = set(new_elb.Subnets)
        subnets_to_add = new_subnets - old_subnets
        subnets_to_remove = old_subnets - new_subnets
        if subnets_to_add:
            manager.attach_to_subnets(
                LoadBalancerName=new_elb.LoadBalancerName, Subnets=list(subnets_to_add)
            )
        if subnets_to_remove:
            manager.detach_from_subnets(
                LoadBalancerName=new_elb.LoadBalancerName,
                Subnets=list(subnets_to_remove),
            )

    def update_availability_zones(
        self, old_elb: "ClassicELB", new_elb: "ClassicELB"
    ) -> None:
        """
        Update the availability zones on an existing :py:class:`ClassicELB` object.

        Args:
            old_elb: The existing :py:class:`ClassicELB
            new_elb: The new :py:class:`ClassicELB` object

        """
        manager = self.using(old_elb.session)
        old_azs = set()
        if old_elb.AvailabilityZones:
            old_azs = set(old_elb.AvailabilityZones)
        new_azs = set()
        if new_elb.AvailabilityZones:
            new_azs = set(new_elb.AvailabilityZones)
        azs_to_add = new_azs - old_azs
        azs_to_remove = old_azs - new_azs
        if azs_to_add:
            manager.enable_availability_zones(
                LoadBalancerName=new_elb.LoadBalancerName,
                AvailabilityZones=list(azs_to_add),
            )
        if azs_to_remove:
            manager.disable_availability_zones(
                LoadBalancerName=new_elb.LoadBalancerName,
                AvailabilityZones=list(azs_to_remove),
            )

    def update_health_check(self, old_elb: "ClassicELB", new_elb: "ClassicELB") -> None:
        """
        Update the health check on an existing :py:class:`ClassicELB` object.

        Args:
            old_elb: The existing :py:class:`ClassicELB
            new_elb: The new :py:class:`ClassicELB` object

        """
        if old_elb.HealthCheck != new_elb.HealthCheck:
            if old_elb.HealthCheck and not new_elb.HealthCheck:
                self.client.configure_health_check(
                    LoadBalancerName=new_elb.LoadBalancerName, HealthCheck={}
                )
            elif (
                not old_elb.HealthCheck and new_elb.HealthCheck
            ) or old_elb.HealthCheck != new_elb.HealthCheck:
                self.configure_health_check(
                    LoadBalancerName=new_elb.LoadBalancerName,
                    HealthCheck=self.serialize(new_elb.HealthCheck),
                )

    def warn_about_policy_updates(
        self, old_elb: "ClassicELB", new_elb: "ClassicELB"
    ) -> None:
        """
        Warn about the fact that:

        * AppCookieStickinessPolicies cannot be updated
        * LBCookieStickinessPolicies cannot be updated
        * OtherPolicies must be updated via :py:meth:`ClassicELB.add_policy` or
          :py:meth:`ClassicELB.delete_policy`.

        Args:
            old_elb: The existing :py:class:`ClassicELB
            new_elb: The new :py:class:`ClassicELB` object

        """
        old_app_has_stickiness = (
            old_elb.Policies and old_elb.Policies.AppCookieStickinessPolicies
        )
        new_app_has_stickiness = (
            new_elb.Policies and new_elb.Policies.AppCookieStickinessPolicies
        )
        if (
            (old_app_has_stickiness and not new_app_has_stickiness)
            or (not old_app_has_stickiness and new_app_has_stickiness)
            or (
                old_app_has_stickiness
                and new_app_has_stickiness
                and old_elb.Policies.AppCookieStickinessPolicies  # type: ignore[union-attr]
                != new_elb.Policies.AppCookieStickinessPolicies  # type: ignore[union-attr]
            )
        ):
            warnings.warn(
                f"ClassicELB(LoadBalancerName='{new_elb.LoadBalancerName}'): "
                "AppCookieStickinessPolicies cannot be updated.  You must delete "
                "and recreate the ELB.",
                UserWarning,
                stacklevel=2,
            )

        old_lb_has_stickiness = (
            old_elb.Policies and old_elb.Policies.LBCookieStickinessPolicies
        )
        new_lb_has_stickiness = (
            new_elb.Policies and new_elb.Policies.LBCookieStickinessPolicies
        )
        if (
            (old_lb_has_stickiness and not new_lb_has_stickiness)
            or (not old_lb_has_stickiness and new_lb_has_stickiness)
            or (
                old_lb_has_stickiness
                and new_lb_has_stickiness
                and old_elb.Policies.LBCookieStickinessPolicies  # type: ignore[union-attr]
                != new_elb.Policies.LBCookieStickinessPolicies  # type: ignore[union-attr]
            )
        ):
            warnings.warn(
                f"ClassicELB(LoadBalancerName='{new_elb.LoadBalancerName}'): "
                "AppCookieStickinessPolicies cannot be updated.  You must delete "
                "and recreate the ELB.",
                UserWarning,
                stacklevel=2,
            )

        old_other_policies = old_elb.Policies.OtherPolicies if old_elb.Policies else []
        new_other_policies = new_elb.Policies.OtherPolicies if new_elb.Policies else []
        if old_other_policies != new_other_policies:
            warnings.warn(
                f"ClassicELB(LoadBalancerName='{new_elb.LoadBalancerName}'): "
                "To change elb.Policies.OtherPolicies, you must use elb.add_policy() "
                "or elb.delete_policy().",
                UserWarning,
                stacklevel=2,
            )

    def update(
        self,
        elb: "ClassicELB",
    ) -> "ClassicELB":
        """
        Update an existing :py:class:`ClassicELB` object.  The reason we have to
        do this in a mixin instead of having this as a usual ``.create()``
        method automatically generated is because Classic ELBs are fragmented
        across multiple API calls.

        We want to pretend that the Classic ELB is a single object, so we have to
        deal with the fragmentation here.

        Args:
            elb: The :py:class:`ClassicELB` object to create

        Returns:
            The newly created :py:class:`ClassicELB` object

        """
        # First get what we have in AWS
        _elb = self.get(elb.LoadBalancerName)

        self.update_tags(_elb, elb)
        self.update_subnets(_elb, elb)
        self.update_availability_zones(_elb, elb)
        self.update_health_check(_elb, elb)
        self.update_attributes(elb)

        # Deal with security groups
        if _elb.SecurityGroups != elb.SecurityGroups:
            security_groups = elb.SecurityGroups if elb.SecurityGroups else []
            elb.apply_security_groups(security_groups)

        # Scheme
        if _elb.Scheme != elb.Scheme:
            warnings.warn(
                f"ClassicELB(LoadBalancerName='{elb.LoadBalancerName}'): "
                "The Scheme cannot be updated.  You must delete and recreate the ELB.",
                UserWarning,
                stacklevel=2,
            )

        self.warn_about_policy_updates(_elb, elb)

        # Listeners are frozen, so we don't have to worry about them
        # Instances are frozen, so we don't have to worry about them
        # SourceSecurityGroup is frozen, so we don't have to worry about it
        # VPCId is frozen so we don't have to worry about it

        return self.get(elb.LoadBalancerName)


class ClassicELBModelMixin:
    """
    Provide a few utility methods for building non-trivial relations.
    """

    @property
    def ec2_instances(self) -> "PrimaryBoto3ModelQuerySet":
        """
        List all the :py:class:`Instance` objects that are part of this ELB.
        """
        # We have to do the actual import here to avoid circular imports
        from botocraft.services import Instance

        return PrimaryBoto3ModelQuerySet(
            [
                Instance.objects.using(self.session).get(instance_id)
                for instance_id in self.Instances
            ]
        )  # type: ignore[arg-type]



import socket
import subprocess
import time
from contextlib import contextmanager
from datetime import datetime
from functools import wraps
from ipaddress import ip_address
from typing import TYPE_CHECKING, Callable, List, Literal, Optional, cast
from zoneinfo import ZoneInfo

import psutil

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.services import (
        AMI,
        Finding,
        Instance,
        Reservation,
        Tag,
        TagSpecification,
    )


#: The EC2 resource types.  We need this for specifying the proper tag for
#: resources in EC2.
ResourceType = Literal[
    "capacity-reservation",
    "client-vpn-endpoint",
    "customer-gateway",
    "carrier-gateway",
    "coip-pool",
    "dedicated-host",
    "dhcp-options",
    "egress-only-internet-gateway",
    "elastic-ip",
    "elastic-gpu",
    "export-image-task",
    "export-instance-task",
    "fleet",
    "fpga-image",
    "host-reservation",
    "image",
    "import-image-task",
    "import-snapshot-task",
    "instance",
    "instance-event-window",
    "internet-gateway",
    "ipam",
    "ipam-pool",
    "ipam-scope",
    "ipv4pool-ec2",
    "ipv6pool-ec2",
    "key-pair",
    "launch-template",
    "local-gateway",
    "local-gateway-route-table",
    "local-gateway-virtual-interface",
    "local-gateway-virtual-interface-group",
    "local-gateway-route-table-vpc-association",
    "local-gateway-route-table-virtual-interface-group-association",
    "natgateway",
    "network-acl",
    "network-interface",
    "network-insights-analysis",
    "network-insights-path",
    "network-insights-access-scope",
    "network-insights-access-scope-analysis",
    "placement-group",
    "prefix-list",
    "replace-root-volume-task",
    "reserved-instances",
    "route-table",
    "security-group",
    "security-group-rule",
    "snapshot",
    "spot-fleet-request",
    "spot-instances-request",
    "subnet",
    "subnet-cidr-reservation",
    "traffic-mirror-filter",
    "traffic-mirror-session",
    "traffic-mirror-target",
    "transit-gateway",
    "transit-gateway-attachment",
    "transit-gateway-connect-peer",
    "transit-gateway-multicast-domain",
    "transit-gateway-policy-table",
    "transit-gateway-route-table",
    "transit-gateway-route-table-announcement",
    "volume",
    "vpc",
    "vpc-endpoint",
    "vpc-endpoint-connection",
    "vpc-endpoint-service",
    "vpc-endpoint-service-permission",
    "vpc-peering-connection",
    "vpn-connection",
    "vpn-gateway",
    "vpc-flow-log",
    "capacity-reservation-fleet",
    "traffic-mirror-filter-rule",
    "vpc-endpoint-connection-device-type",
    "verified-access-instance",
    "verified-access-group",
    "verified-access-endpoint",
    "verified-access-policy",
    "verified-access-trust-provider",
    "vpn-connection-device-type",
    "vpc-block-public-access-exclusion",
    "ipam-resource-discovery",
    "ipam-resource-discovery-association",
    "instance-connect-endpoint",
]

# ----------
# Decorators
# ----------


def ec2_instances_only(
    func: Callable[..., "PrimaryBoto3ModelQuerySet"],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps a boto3 method that returns a list of :py:class:`Reservation` objects
    to return a list of :py:class:`Instance` objects instead.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        qs = func(*args, **kwargs)
        instances: List["Instance"] = []  # noqa: UP037
        if hasattr(qs, "results") and qs.results:
            for reservation in qs.results:
                instances.extend(cast("List[Instance]", reservation.Instances))  # type: ignore[attr-defined]
            return PrimaryBoto3ModelQuerySet(instances)  # type: ignore[arg-type]
        return PrimaryBoto3ModelQuerySet([])

    return wrapper


def ec2_instance_only(
    func: Callable[..., Optional["Reservation"]],
) -> Callable[..., Optional["Instance"]]:
    """
    Wraps a boto3 method that returns a list of :py:class:`Reservation` objects
    to return a single :py:class:`Instance` object instead.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> Optional["Instance"]:
        reservation = func(*args, **kwargs)
        if not reservation:
            return None
        return cast("List[Instance]", reservation.Instances)[0]

    return wrapper


# --------------
# Manager mixins
# --------------


class EC2TagsManagerMixin:
    """
    A mixin is used on on :py:class:`botocraft.services.ec2.InstanceManager`
    to convert the odd EC2 tag list to a :py:class:`TagSpecification` object.
    """

    def convert_tags(
        self, tags: List["Tag"] | None, resource_type: ResourceType
    ) -> Optional["TagSpecification"]:
        """
        Given a TagList, convert it to a TagSpecification with ResourceType of
        ``resource_type``.

        Args:
            tags: the list of :py:class:`Tag` objects to convert.
            resource_type: the EC2 resource type.

        Returns:
            A :py:class:`TagSpecification` object, or ``None`` if ``tags`` is
            ``None``.

        """
        from botocraft.services import TagSpecification

        if tags is None:
            return None
        return TagSpecification(ResourceType=resource_type, Tags=tags)


class AMIManagerMixin:
    """
    A mixin is used on :py:class:`botocraft.services.ec2.AMIManager` to add
    miscellaneous methods to the class that are not normally part of the
    object.
    """

    def _get_in_use_instance_amis(
        self,
        check_amis: list["AMI"],
        amis: list["AMI"] | None = None,
    ) -> list["AMI"]:
        """
        Return a list of AMIs that are in use by a running or stopped instance.

        Args:
            check_amis: the list of :py:class:`AMI` to check.

        Keyword Args:
            amis: the list of :py:class:`AMI` that have been identified so far as
              being in use.

        Returns:
            A list of :py:class:`AMI` objects that are in use by a running or
            stopped instance.

        """
        from botocraft.services import (
            Filter,
            Instance,
        )

        if not amis:
            amis = []
        found_amis: list[str] = Instance.objects.list(
            Filters=[
                Filter(Name="image-id", Values=[ami.ImageId for ami in check_amis])
            ]
        ).values_list("ImageId", flat=True)
        amis.extend(
            [ami for ami in check_amis if ami.ImageId in found_amis and ami not in amis]
        )
        return amis

    def _get_in_use_asg_amis(
        self,
        check_amis: List["AMI"],
        amis: List["AMI"] | None = None,
    ) -> List["AMI"]:
        """
        Return a list of AMIs that are in use by an autoscaling group.
        """
        # Avoid circular import
        from botocraft.services import (
            AutoScalingGroup,
            LaunchConfiguration,
            LaunchTemplateVersion,
            ResponseLaunchTemplateData,
        )

        if amis is None:
            amis = []
        # Now search for any AMIs that are used by an autoscaling group
        autoscaling_groups = AutoScalingGroup.objects.list()
        for autoscaling_group in autoscaling_groups:
            autoscaling_group = cast("AutoScalingGroup", autoscaling_group)
            for ami in check_amis:
                # First check if the AMI is used by a launch configuration
                if autoscaling_group.LaunchConfigurationName:
                    if (
                        cast(
                            "LaunchConfiguration",
                            autoscaling_group.launch_configuration,
                        ).ImageId
                        == ami.ImageId
                    ):
                        if ami not in amis:
                            amis.append(ami)
                    continue
                # If there's no launch configuration, then the ASG uses a launch
                # template.  Check if the AMI is used by the launch template.
                template = cast(
                    "LaunchTemplateVersion", autoscaling_group.launch_template
                )
                if template:
                    if (
                        cast(
                            "ResponseLaunchTemplateData", template.LaunchTemplateData
                        ).ImageId
                        == ami.ImageId
                    ):
                        if ami not in amis:
                            amis.append(ami)
        return amis

    def in_use(
        self,
        owners: list[str] | None = None,
        tags: dict[str, str] | None = None,
        created_since: datetime | None = None,
        amis: list["AMI"] | None = None,
    ) -> "PrimaryBoto3ModelQuerySet":
        """
        Return a list of AMIs that are currently in use by a running or stopped
        instance, or by an autoscaling group's launch configuration or launch
        template.

        .. important::

            We're not checking for AMIs in the following places:

            - CloudFormation templates
            - OpsWorks stacks
            - Elastic Beanstalk environments
            - Launch Template Versions not in use by an autoscaling group

        Keyword Args:
            owners: Scopes the results to AMIs with the specified owners. You
                can specify a combination of Amazon Web Services account IDs,
                ``self``, ``amazon``, and ``aws-marketplace``. If you omit this
                parameter, the results include all images for which you have launch
                permissions, regardless of ownership.  If not specified, the
                default is ``self``.
            tags: Filters the AMIs to those who match the these tags.
            created_since: Filters the AMIs to those created since this date.
            amis: Filters the AMIs to those in this list.  All other filters are
                ignored if this is specified.

        """
        from botocraft.services import (
            Filter,
        )

        _owners = owners if owners else ["self"]
        _filters: list[Filter] = []
        if tags:
            _filters = [
                Filter(Name=f"tag:{key}", Values=[value]) for key, value in tags.items()
            ]
        if created_since:
            # First convert the timezone to UTC if this is a timezone-aware
            # datetime object.
            if created_since.tzinfo:
                created_since = created_since.astimezone(ZoneInfo("UTC"))
            # Now append the filter to the list of filters for the AMI listing.
            _filters.append(
                Filter(Name="creation-date", Values=[created_since.isoformat()])
            )
        if amis is None:
            check_amis = self.list(Owners=_owners, Filters=_filters)  # type: ignore[attr-defined]
        else:
            check_amis = amis

        in_use_amis = self._get_in_use_instance_amis(check_amis)
        in_use_amis = self._get_in_use_asg_amis(check_amis, amis=in_use_amis)
        return PrimaryBoto3ModelQuerySet(in_use_amis)  # type: ignore[arg-type]


# -------------
# Model mixins
# -------------


class InstanceModelMixin:
    """
    Used on :py:class:`botocraft.services.ec2.Instance` to add miscellaneous
    methods to the class that are not normally part of the object.
    """

    def __maybe_resolve_ip(self, host: str) -> str:
        """
        Given a hostname or IP address, return the IP address.  If the host is
        an IP address, return it as is.  If the host is a hostname, resolve it
        to an IP address.

        Args:
            host: either an IP address of the remote host to connect to, or
                a hostname of that host

        Returns:
            The IP address of the host.

        """
        try:
            ip_address(host)
        except ValueError:
            # If it is not an IP address, then it must be a hostname.
            # Resolve the hostname to an IP address.
            try:
                return socket.gethostbyname(host)
            except socket.gaierror as e:
                msg = f"Could not resolve hostname {host}: {e}"
                raise RuntimeError(msg) from e
        return host

    def __find_open_port(self, start_port: int) -> int:
        """
        Find an open port starting from ``start_port``.  This is used to find
        an open port for the SSH tunnel.

        Args:
            start_port: The port to start searching from.

        Returns:
            An open port.

        Raises:
            ValueError: If no open port is found.

        """
        port = start_port
        while True:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                if sock.connect_ex(("127.0.0.1", port)) != 0:
                    break
                port += 1
                if port > 65535:  # noqa: PLR2004
                    msg = f"No available port found starting at {start_port}."
                    raise ValueError(msg)
        return port

    @contextmanager
    def tunnel(
        self,
        host: str,
        remote_port: int,
        local_port: int | None = None,
        profile: str | None = None,
    ):
        """
        A context manager for opening and closing a tunnel.

        This will call `open_tunnel` when entering the context and
        `close_tunnel` when exiting the context.

        Args:
            host: The remote host to connect to (IP or hostname).
            remote_port: The remote port to connect to.
            local_port: The local port to use. If None, an unused port will be chosen.
            profile: The AWS profile to use. If None, the default profile will be used.

        Raises:
            ValueError: If the local port is already in use.
            RuntimeError: If the instance is not connected to SSM or if there
                is an error when starting the session or SSH tunnel.

        Yields:
            The local port used for the tunnel.

        """
        local_port = self.open_tunnel(
            host=host,
            remote_port=remote_port,
            local_port=local_port,
            profile=profile,
        )
        try:
            yield local_port
        finally:
            self.close_tunnel(host=host, local_port=local_port)

    def open_tunnel(
        self,
        host: str,
        remote_port: int,
        local_port: int | None = None,
        profile: str | None = None,
    ) -> int:
        """
        Open a tunnel to the instance using SSM and SSH. This is useful for
        connecting to a database or other service on the instance that
        is in a private subnet. This will open a tunnel from the ``local_port``
        on the local machine through the instance to the ``remote_port`` on
        ``host``.

        If ``local_port`` is not specified, a random port will be chosen
        starting from between 8800 and 65535.

        Args:
            host: either an IP address of the remote host to connect to, or
                a hostname of that host
            remote_port: The remote port to connect to.

        Keyword Args:
            local_port: The local port to connect to.
            profile: The AWS profile to use. If not specified, the default
                profile will be used.

        Raises:
            ValueError: If the local port is already in use.
            RuntimeError: If the instance is not connected to SSM or if there
                is an error when starting the session or SSH tunnel.

        Returns:
            The local port that was used for the tunnel.

        """
        # Find an unused local port if local_port is None
        if local_port is None:
            local_port = self.__find_open_port(8800)
        else:
            # Check if the local port is already in use
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                if sock.connect_ex(("127.0.0.1", local_port)) == 0:
                    msg = f"Local port {local_port} is already in use."
                    raise ValueError(msg)
        host_ip = self.__maybe_resolve_ip(host)

        # Build the AWS SSM start-session command
        ssm_command = [
            "aws",
            "ssm",
            "start-session",
            "--target",
            self.InstanceId,  # type: ignore[attr-defined]
            "--document-name",
            "AWS-StartPortForwardingSessionToRemoteHost",
            "--parameters",
            f"host={host_ip},portNumber={remote_port},localPortNumber={local_port}",
        ]

        if profile:
            ssm_command.extend(["--profile", profile])

        try:
            # Start the SSM session
            ssm_process = subprocess.Popen(
                ssm_command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except FileNotFoundError as e:
            msg = "AWS CLI is not installed or not in PATH"
            raise RuntimeError(msg) from e

        # Wait for the port to be open
        for _ in range(40):
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                if sock.connect_ex(("127.0.0.1", local_port)) == 0:
                    break
            time.sleep(0.25)
        else:
            # If the port is not open after 10 seconds, kill the process
            ssm_process.kill()
            msg = (
                f"Failed to open tunnel to {host_ip}:{remote_port} on local "
                f"port {local_port}."
            )
            raise RuntimeError(msg)

        # Store the processes for later management
        if self.Tunnels is None:  # type: ignore[has-type]
            self.Tunnels = {}  # type: ignore[var-annotated]
        if host_ip not in self.Tunnels:
            self.Tunnels[host_ip] = []
        self.Tunnels[host_ip].append(
            {
                "ssm_process": ssm_process,
                "local_port": local_port,
            }
        )
        return local_port

    def close_tunnel(self, host: str, local_port: int | None = None) -> None:
        """
        Close one or more tunnels to ``host``. This will terminate the SSM
        session(s) and SSH process(es) that were opened by
        :py:meth:`start_tunnel`.

        If ``local_port`` is not specified, all tunnels to the host will be
        closed. If ``local_port`` is specified, only the tunnel to that port
        will be closed.

        Args:
            host: Either an IP address of the remote host to connect to, or
                a hostname of that host.

        Keyword Args:
            local_port: The local port to close. If not specified, all
                tunnels to the host will be closed.

        Raises:
            ValueError: If no tunnels are found for the given host or port.

        """

        def terminate_process(process: subprocess.Popen) -> None:
            """
            Helper function to terminate a process and its children.

            Args:
                process: The process to terminate.

            """
            try:
                for child in psutil.Process(process.pid).children(recursive=True):
                    child.terminate()
                process.terminate()
            except psutil.NoSuchProcess:
                pass

        host_ip = self.__maybe_resolve_ip(host)

        if self.Tunnels is None or host_ip not in self.Tunnels:
            msg = f"No open tunnels found for {host_ip}. Did you call start_tunnel?"
            raise ValueError(msg)

        tunnels = self.Tunnels[host_ip]

        if local_port is not None:
            # Close the specific tunnel for the given local port
            for tunnel in tunnels:
                if tunnel["local_port"] == local_port:
                    terminate_process(tunnel["ssm_process"])
                    tunnels.remove(tunnel)
                    break
            else:
                msg = f"No open tunnel found for {host_ip} on local port {local_port}."
                raise ValueError(msg)
        else:
            # Close all tunnels for the given host
            for tunnel in tunnels:
                terminate_process(tunnel["ssm_process"])
            del self.Tunnels[host_ip]


class SecurityGroupModelMixin:
    """
    A mixin is used on :py:class:`botocraft.services.ec2.SecurityGroup` to
    enhance the ``.save()`` method to allow for managing ingress and egress
    rules at the same time as saving the security group.

    Normally this is done with several boto3 calls, but this mixin allows for a
    single call to ``.save()`` to create a security group and manage the rules.
    """

    def save(self, **kwargs):
        """
        Save the model.  For security groups, ingress rules are managed via
        separate boto3 calls than the security group itself.  This override of
        the ``save`` method will allow the user to create a security group and
        add ingress rules in one step.
        """
        # TODO: this needs to be enhanced to handle egress rules as well.
        if not self.pk:
            group_id = self.objects.create(self, **kwargs)
            self.objects.using(self.session).authorize_ingress(
                group_id, self.IpPermissions, **kwargs
            )
        else:
            old_obj = self.objects.using(self.session).get(self.pk)
            if self.IpPermissions != old_obj.IpPermissions:
                if old_obj.IpPermissions:
                    self.objects.using(self.session).revoke_ingress(
                        self.pk, old_obj.IpPermissions, **kwargs
                    )
                if self.IpPermissions:
                    self.objects.using(self.session).authorize_ingress(
                        self.pk, self.IpPermissions, **kwargs
                    )


class AMIModelMixin:
    @property
    def in_use(self) -> bool:
        """
        Return ``True`` if the AMI is in use by a running or stopped instance.
        """
        ids = self.objects.in_use(image_id=self.ImageId).values_list(  # type: ignore[attr-defined]
            "ImageId", flat=True
        )
        return self.ImageId in ids  # type: ignore[attr-defined]

    @property
    def vulnerabilities(self) -> "PrimaryBoto3ModelQuerySet":
        """
        Return a list of vulnerabilities for the instance.  This is a
        convenience method to get the vulnerabilities for the instance.

        Returns:
            A list of :py:class:`Finding` objects.

        """
        from botocraft.services import (
            FilterCriteria,
            Finding,
            StringFilter,
        )

        return Finding.objects.using(self.session).list(  # type: ignore[attr-defined]
            FilterCriteria(
                ec2InstanceImageId=[
                    StringFilter(value=self.ImageId, comparison="EQUALS")  # type: ignore[attr-defined]
                ],
            ),
        )


class SubnetModelMixin:
    """
    A mixin is used on :py:class:`botocraft.services.ec2.Subnet` to add
    miscellaneous methods to the class that are not normally part of the
    object.
    """

    @property
    def vulnerabilities(self) -> List["Finding"]:
        """
        Return a list of vulnerabilities for the instance.  This is a
        convenience method to get the vulnerabilities for the instance.

        Returns:
            A list of :py:class:`Finding` objects.

        """
        from botocraft.services import (
            FilterCriteria,
            Finding,
            StringFilter,
        )

        return Finding.objects.using(self.session).list(  # type: ignore[attr-defined]
            filterCriteria=FilterCriteria(
                ec2InstanceSubnetId=[
                    StringFilter(value=self.SubnetId, comparison="EQUALS")  # type: ignore[attr-defined]
                ],
            ),
        )


class VpcModelMixin:
    """
    A mixin is used on :py:class:`botocraft.services.ec2.Vpc` to add
    miscellaneous methods to the class that are not normally part of the
    object.
    """

    @property
    def vulnerabilities(self) -> List["Finding"]:
        """
        Return a list of vulnerabilities for the instance.  This is a
        convenience method to get the vulnerabilities for the instance.

        Returns:
            A list of :py:class:`Finding` objects.

        """
        from botocraft.services import (
            FilterCriteria,
            Finding,
            StringFilter,
        )

        return Finding.objects.using(self.session).list(  # type: ignore[attr-defined]
            filterCriteria=FilterCriteria(
                ec2InstanceVpcId=[
                    StringFilter(value=self.VpcId, comparison="EQUALS")  # type: ignore[attr-defined]
                ],
            ),
        )



# mypy: disable-error-code="attr-defined"
from functools import wraps
from typing import TYPE_CHECKING, Callable, List, Optional

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.services import DescribeRuleResponse, EventBus, EventRule


def event_rules_only(
    func: Callable[..., List[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.events.EventRuleManager.list_by_target`
    to return :py:class:`botocraft.services.ecs.EventRule` objects instead of
    only a list of Rule names.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        names = func(self, *args, **kwargs)
        events = [self.get(name) for name in names]
        return PrimaryBoto3ModelQuerySet(events)

    return wrapper


def DescribeRuleResponse_to_EventRule(
    func: Callable[..., Optional["DescribeRuleResponse"]],
) -> Callable[..., Optional["EventRule"]]:
    """
    The boto3 call describe_rule does not actually return a rule object, but
    instead a DescribeRuleResponse object. This decorator wraps the function
    to return a Rule object instead.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Optional["EventRule"]:
        from botocraft.services import EventRule

        response = func(self, *args, **kwargs)
        if not response:
            return None
        rule = EventRule(**response.model_dump())
        self.sessionize(rule)
        return rule

    return wrapper


def EventRule_purge_CreatedBy_attribute(
    func: Callable[..., str],
) -> Callable[..., str]:
    """
    The boto3 call describe_rule does not actually return a rule object, but
    instead a DescribeRuleResponse object. This decorator wraps the function
    to return a Rule object instead.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> str:
        model = args[0]
        model.CreatedBy = None
        return func(self, *args, **kwargs)

    return wrapper


def DescribeEventBusResponse_to_EventBus(
    func: Callable[..., Optional["DescribeRuleResponse"]],
) -> Callable[..., Optional["EventBus"]]:
    """
    The boto3 call describe_rule does not actually return a rule object, but
    instead a DescribeRuleResponse object. This decorator wraps the function
    to return a Rule object instead.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> Optional["EventRule"]:
        from botocraft.services.events import EventBus

        response = func(self, *args, **kwargs)
        if not response:
            return None
        bus = EventBus(**response.model_dump())
        self.sessionize(bus)
        return bus  # type: ignore[return-value]

    return wrapper  # type: ignore[return-value]



# This file is automatically generated by botocraft.  Do not edit directly.
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
from botocraft.mixins.tags import TagsDictMixin
from botocraft.services.common import Tag
from .abstract import (
    Boto3Model,
    ReadonlyBoto3Model,
    PrimaryBoto3Model,
    ReadonlyPrimaryBoto3Model,
    Boto3ModelManager,
    ReadonlyBoto3ModelManager,
)
from botocraft.services.ec2 import Vpc, VpcManager
from typing import ClassVar, Type, Optional, Literal, Dict, List, Union, Any, cast
from botocraft.services.ec2 import SecurityGroup, SecurityGroupManager
from functools import cached_property
from botocraft.services.common import Filter
from .abstract import PrimaryBoto3ModelQuerySet
from datetime import datetime
from collections import OrderedDict
from pydantic import Field

# ===============
# Managers
# ===============


class DBInstanceManager(Boto3ModelManager):
    service_name: str = "rds"

    def create(
        self,
        model: "DBInstance",
        MasterUserPassword: str | None = None,
        VpcSecurityGroupIds: List[str] | None = None,
        DBSubnetGroupName: str | None = None,
        DBParameterGroupName: str | None = None,
        OptionGroupName: str | None = None,
        TdeCredentialPassword: str | None = None,
        Domain: str | None = None,
        DomainFqdn: str | None = None,
        DomainOu: str | None = None,
        DomainAuthSecretArn: str | None = None,
        DomainDnsIps: List[str] | None = None,
        DomainIAMRoleName: str | None = None,
        ManageMasterUserPassword: bool | None = None,
        MasterUserSecretKmsKeyId: str | None = None,
    ) -> "DBInstance":
        """
        Creates a new DB instance.

        Args:
            model: The :py:class:`DBInstance` to create.

        Keyword Args:
            MasterUserPassword: The password for the master user.
            VpcSecurityGroupIds: A list of Amazon EC2 VPC security groups to associate with this DB instance.
            DBSubnetGroupName: A DB subnet group to associate with this DB instance.
            DBParameterGroupName: The name of the DB parameter group to associate with this DB instance. If you don't specify a
                value, then Amazon RDS uses the default DB parameter group for the specified DB engine and version.
            OptionGroupName: The option group to associate the DB instance with.
            TdeCredentialPassword: The password for the given ARN from the key store in order to access the device.
            Domain: The Active Directory directory ID to create the DB instance in. Currently, you can create only Db2, MySQL,
                Microsoft SQL Server, Oracle, and PostgreSQL DB instances in an Active Directory Domain.
            DomainFqdn: The fully qualified domain name (FQDN) of an Active Directory domain.
            DomainOu: The Active Directory organizational unit for your DB instance to join.
            DomainAuthSecretArn: The ARN for the Secrets Manager secret with the credentials for the user joining the domain.
            DomainDnsIps: The IPv4 DNS IP addresses of your primary and secondary Active Directory domain controllers.
            DomainIAMRoleName: The name of the IAM role to use when making API calls to the Directory Service.
            ManageMasterUserPassword: Specifies whether to manage the master user password with Amazon Web Services Secrets
                Manager.
            MasterUserSecretKmsKeyId: The Amazon Web Services KMS key identifier to encrypt a secret that is automatically
                generated and managed in Amazon Web Services Secrets Manager.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            DBInstanceIdentifier=data.get("DBInstanceIdentifier"),
            DBInstanceClass=data.get("DBInstanceClass"),
            Engine=data.get("Engine"),
            DBName=data.get("DBName"),
            AllocatedStorage=data.get("AllocatedStorage"),
            MasterUsername=data.get("MasterUsername"),
            MasterUserPassword=self.serialize(MasterUserPassword),
            VpcSecurityGroupIds=self.serialize(VpcSecurityGroupIds),
            AvailabilityZone=data.get("AvailabilityZone"),
            DBSubnetGroupName=self.serialize(DBSubnetGroupName),
            PreferredMaintenanceWindow=data.get("PreferredMaintenanceWindow"),
            DBParameterGroupName=self.serialize(DBParameterGroupName),
            BackupRetentionPeriod=data.get("BackupRetentionPeriod"),
            PreferredBackupWindow=data.get("PreferredBackupWindow"),
            Port=data.get("DbInstancePort"),
            MultiAZ=data.get("MultiAZ"),
            EngineVersion=data.get("EngineVersion"),
            AutoMinorVersionUpgrade=data.get("AutoMinorVersionUpgrade"),
            LicenseModel=data.get("LicenseModel"),
            Iops=data.get("Iops"),
            OptionGroupName=self.serialize(OptionGroupName),
            CharacterSetName=data.get("CharacterSetName"),
            NcharCharacterSetName=data.get("NcharCharacterSetName"),
            PubliclyAccessible=data.get("PubliclyAccessible"),
            Tags=data.get("TagList"),
            DBClusterIdentifier=data.get("DBClusterIdentifier"),
            StorageType=data.get("StorageType"),
            TdeCredentialArn=data.get("TdeCredentialArn"),
            TdeCredentialPassword=self.serialize(TdeCredentialPassword),
            StorageEncrypted=data.get("StorageEncrypted"),
            KmsKeyId=data.get("KmsKeyId"),
            Domain=self.serialize(Domain),
            DomainFqdn=self.serialize(DomainFqdn),
            DomainOu=self.serialize(DomainOu),
            DomainAuthSecretArn=self.serialize(DomainAuthSecretArn),
            DomainDnsIps=self.serialize(DomainDnsIps),
            CopyTagsToSnapshot=data.get("CopyTagsToSnapshot"),
            MonitoringInterval=data.get("MonitoringInterval"),
            MonitoringRoleArn=data.get("MonitoringRoleArn"),
            DomainIAMRoleName=self.serialize(DomainIAMRoleName),
            PromotionTier=data.get("PromotionTier"),
            Timezone=data.get("Timezone"),
            EnableIAMDatabaseAuthentication=data.get(
                "IAMDatabaseAuthenticationEnabled"
            ),
            DatabaseInsightsMode=data.get("DatabaseInsightsMode"),
            EnablePerformanceInsights=data.get("PerformanceInsightsEnabled"),
            PerformanceInsightsKMSKeyId=data.get("PerformanceInsightsKMSKeyId"),
            PerformanceInsightsRetentionPeriod=data.get(
                "PerformanceInsightsRetentionPeriod"
            ),
            EnableCloudwatchLogsExports=data.get("EnabledCloudwatchLogsExports"),
            ProcessorFeatures=data.get("ProcessorFeatures"),
            DeletionProtection=data.get("DeletionProtection"),
            MaxAllocatedStorage=data.get("MaxAllocatedStorage"),
            EnableCustomerOwnedIp=data.get("CustomerOwnedIpEnabled"),
            CustomIamInstanceProfile=data.get("CustomIamInstanceProfile"),
            BackupTarget=data.get("BackupTarget"),
            NetworkType=data.get("NetworkType"),
            StorageThroughput=data.get("StorageThroughput"),
            ManageMasterUserPassword=self.serialize(ManageMasterUserPassword),
            MasterUserSecretKmsKeyId=self.serialize(MasterUserSecretKmsKeyId),
            CACertificateIdentifier=data.get("CACertificateIdentifier"),
            DBSystemId=data.get("DBSystemId"),
            DedicatedLogVolume=data.get("DedicatedLogVolume"),
            MultiTenant=data.get("MultiTenant"),
            EngineLifecycleSupport=data.get("EngineLifecycleSupport"),
        )
        _response = self.client.create_db_instance(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateDBInstanceResult(**_response)

        self.sessionize(response.RDSDBInstance)
        return cast("DBInstance", response.RDSDBInstance)

    def update(
        self,
        model: "DBInstance",
        DBSubnetGroupName: str | None = None,
        VpcSecurityGroupIds: List[str] | None = None,
        ApplyImmediately: bool | None = None,
        MasterUserPassword: str | None = None,
        DBParameterGroupName: str | None = None,
        AllowMajorVersionUpgrade: bool | None = None,
        OptionGroupName: str | None = None,
        NewDBInstanceIdentifier: str | None = None,
        TdeCredentialPassword: str | None = None,
        Domain: str | None = None,
        DomainFqdn: str | None = None,
        DomainOu: str | None = None,
        DomainAuthSecretArn: str | None = None,
        DomainDnsIps: List[str] | None = None,
        DBPortNumber: int | None = None,
        DomainIAMRoleName: str | None = None,
        DisableDomain: bool | None = None,
        CloudwatchLogsExportConfiguration: Optional[
            "CloudwatchLogsExportConfiguration"
        ] = None,
        UseDefaultProcessorFeatures: bool | None = None,
        CertificateRotationRestart: bool | None = None,
        ResumeFullAutomationModeMinutes: int | None = None,
        ManageMasterUserPassword: bool | None = None,
        RotateMasterUserPassword: bool | None = None,
        MasterUserSecretKmsKeyId: str | None = None,
    ) -> "DBInstance":
        """
        Modifies settings for a DB instance. You can change one or more database
        configuration parameters by specifying these parameters and the new values in
        the request. To learn what modifications you can make to your DB instance, call
        ``DescribeValidDBInstanceModifications`` before you call ``ModifyDBInstance``.

        Args:
            model: The :py:class:`DBInstance` to update.

        Keyword Args:
            DBSubnetGroupName: The new DB subnet group for the DB instance. You can use this parameter to move your DB instance
                to a different VPC. If your DB instance isn't in a VPC, you can also use this parameter to move your DB instance
                into a VPC. For more information, see `Working with a DB instance in a VPC
                <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/USER_VPC.WorkingWithRDSInstanceinaVPC.html#USER_VPC.Non-
                VPC2VPC>`_ in the *Amazon RDS User Guide*.
            VpcSecurityGroupIds: A list of Amazon EC2 VPC security groups to associate with this DB instance. This change is
                asynchronously applied as soon as possible.
            ApplyImmediately: Specifies whether the modifications in this request and any pending modifications are
                asynchronously applied as soon as possible, regardless of the ``PreferredMaintenanceWindow`` setting for the DB
                instance. By default, this parameter is disabled.
            MasterUserPassword: The new password for the master user.
            DBParameterGroupName: The name of the DB parameter group to apply to the DB instance.
            AllowMajorVersionUpgrade: Specifies whether major version upgrades are allowed. Changing this parameter doesn't
                result in an outage and the change is asynchronously applied as soon as possible.
            OptionGroupName: The option group to associate the DB instance with.
            NewDBInstanceIdentifier: The new identifier for the DB instance when renaming a DB instance. When you change the DB
                instance identifier, an instance reboot occurs immediately if you enable ``ApplyImmediately``, or will occur during
                the next maintenance window if you disable ``ApplyImmediately``. This value is stored as a lowercase string.
            TdeCredentialPassword: The password for the given ARN from the key store in order to access the device.
            Domain: The Active Directory directory ID to move the DB instance to. Specify ``none`` to remove the instance from
                its current domain. You must create the domain before this operation. Currently, you can create only Db2, MySQL,
                Microsoft SQL Server, Oracle, and PostgreSQL DB instances in an Active Directory Domain.
            DomainFqdn: The fully qualified domain name (FQDN) of an Active Directory domain.
            DomainOu: The Active Directory organizational unit for your DB instance to join.
            DomainAuthSecretArn: The ARN for the Secrets Manager secret with the credentials for the user joining the domain.
            DomainDnsIps: The IPv4 DNS IP addresses of your primary and secondary Active Directory domain controllers.
            DBPortNumber: The port number on which the database accepts connections.
            DomainIAMRoleName: The name of the IAM role to use when making API calls to the Directory Service.
            DisableDomain: Specifies whether to remove the DB instance from the Active Directory domain.
            CloudwatchLogsExportConfiguration: The log types to be enabled for export to CloudWatch Logs for a specific DB
                instance.
            UseDefaultProcessorFeatures: Specifies whether the DB instance class of the DB instance uses its default processor
                features.
            CertificateRotationRestart: Specifies whether the DB instance is restarted when you rotate your SSL/TLS certificate.
            ResumeFullAutomationModeMinutes: The number of minutes to pause the automation. When the time period ends, RDS
                Custom resumes full automation.
            ManageMasterUserPassword: Specifies whether to manage the master user password with Amazon Web Services Secrets
                Manager.
            RotateMasterUserPassword: Specifies whether to rotate the secret managed by Amazon Web Services Secrets Manager for
                the master user password.
            MasterUserSecretKmsKeyId: The Amazon Web Services KMS key identifier to encrypt a secret that is automatically
                generated and managed in Amazon Web Services Secrets Manager.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            DBInstanceIdentifier=data.get("DBInstanceIdentifier"),
            AllocatedStorage=data.get("AllocatedStorage"),
            DBInstanceClass=data.get("DBInstanceClass"),
            DBSubnetGroupName=self.serialize(DBSubnetGroupName),
            VpcSecurityGroupIds=self.serialize(VpcSecurityGroupIds),
            ApplyImmediately=self.serialize(ApplyImmediately),
            MasterUserPassword=self.serialize(MasterUserPassword),
            DBParameterGroupName=self.serialize(DBParameterGroupName),
            BackupRetentionPeriod=data.get("BackupRetentionPeriod"),
            PreferredBackupWindow=data.get("PreferredBackupWindow"),
            PreferredMaintenanceWindow=data.get("PreferredMaintenanceWindow"),
            MultiAZ=data.get("MultiAZ"),
            EngineVersion=data.get("EngineVersion"),
            AllowMajorVersionUpgrade=self.serialize(AllowMajorVersionUpgrade),
            AutoMinorVersionUpgrade=data.get("AutoMinorVersionUpgrade"),
            LicenseModel=data.get("LicenseModel"),
            Iops=data.get("Iops"),
            OptionGroupName=self.serialize(OptionGroupName),
            NewDBInstanceIdentifier=self.serialize(NewDBInstanceIdentifier),
            StorageType=data.get("StorageType"),
            TdeCredentialArn=data.get("TdeCredentialArn"),
            TdeCredentialPassword=self.serialize(TdeCredentialPassword),
            CACertificateIdentifier=data.get("CACertificateIdentifier"),
            Domain=self.serialize(Domain),
            DomainFqdn=self.serialize(DomainFqdn),
            DomainOu=self.serialize(DomainOu),
            DomainAuthSecretArn=self.serialize(DomainAuthSecretArn),
            DomainDnsIps=self.serialize(DomainDnsIps),
            CopyTagsToSnapshot=data.get("CopyTagsToSnapshot"),
            MonitoringInterval=data.get("MonitoringInterval"),
            DBPortNumber=self.serialize(DBPortNumber),
            PubliclyAccessible=data.get("PubliclyAccessible"),
            MonitoringRoleArn=data.get("MonitoringRoleArn"),
            DomainIAMRoleName=self.serialize(DomainIAMRoleName),
            DisableDomain=self.serialize(DisableDomain),
            PromotionTier=data.get("PromotionTier"),
            EnableIAMDatabaseAuthentication=data.get(
                "IAMDatabaseAuthenticationEnabled"
            ),
            DatabaseInsightsMode=data.get("DatabaseInsightsMode"),
            EnablePerformanceInsights=data.get("PerformanceInsightsEnabled"),
            PerformanceInsightsKMSKeyId=data.get("PerformanceInsightsKMSKeyId"),
            PerformanceInsightsRetentionPeriod=data.get(
                "PerformanceInsightsRetentionPeriod"
            ),
            CloudwatchLogsExportConfiguration=self.serialize(
                CloudwatchLogsExportConfiguration
            ),
            ProcessorFeatures=data.get("ProcessorFeatures"),
            UseDefaultProcessorFeatures=self.serialize(UseDefaultProcessorFeatures),
            DeletionProtection=data.get("DeletionProtection"),
            MaxAllocatedStorage=data.get("MaxAllocatedStorage"),
            CertificateRotationRestart=self.serialize(CertificateRotationRestart),
            ReplicaMode=data.get("ReplicaMode"),
            EnableCustomerOwnedIp=data.get("CustomerOwnedIpEnabled"),
            AwsBackupRecoveryPointArn=data.get("AwsBackupRecoveryPointArn"),
            AutomationMode=data.get("AutomationMode"),
            ResumeFullAutomationModeMinutes=self.serialize(
                ResumeFullAutomationModeMinutes
            ),
            NetworkType=data.get("NetworkType"),
            StorageThroughput=data.get("StorageThroughput"),
            ManageMasterUserPassword=self.serialize(ManageMasterUserPassword),
            RotateMasterUserPassword=self.serialize(RotateMasterUserPassword),
            MasterUserSecretKmsKeyId=self.serialize(MasterUserSecretKmsKeyId),
            Engine=data.get("Engine"),
            DedicatedLogVolume=data.get("DedicatedLogVolume"),
            MultiTenant=data.get("MultiTenant"),
        )
        _response = self.client.modify_db_instance(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = ModifyDBInstanceResult(**_response)

        self.sessionize(response.RDSDBInstance)
        return cast("DBInstance", response.RDSDBInstance)

    def delete(
        self,
        DBInstanceIdentifier: str,
        *,
        SkipFinalSnapshot: bool | None = None,
        FinalDBSnapshotIdentifier: str | None = None,
        DeleteAutomatedBackups: bool | None = None,
    ) -> "DBInstance":
        """
        Deletes a previously provisioned DB instance. When you delete a DB instance, all
        automated backups for that instance are deleted and can't be recovered. However,
        manual DB snapshots of the DB instance aren't deleted.

        Args:
            DBInstanceIdentifier: The DB instance identifier for the DB instance to be deleted. This parameter isn't case-
                sensitive.

        Keyword Args:
            SkipFinalSnapshot: Specifies whether to skip the creation of a final DB snapshot before deleting the instance. If
                you enable this parameter, RDS doesn't create a DB snapshot. If you don't enable this parameter, RDS creates a DB
                snapshot before the DB instance is deleted. By default, skip isn't enabled, and the DB snapshot is created.
            FinalDBSnapshotIdentifier: The ``DBSnapshotIdentifier`` of the new ``DBSnapshot`` created when the
                ``SkipFinalSnapshot`` parameter is disabled.
            DeleteAutomatedBackups: Specifies whether to remove automated backups immediately after the DB instance is deleted.
                This parameter isn't case- sensitive. The default is to remove automated backups immediately after the DB instance
                is deleted.
        """
        args: Dict[str, Any] = dict(
            DBInstanceIdentifier=self.serialize(DBInstanceIdentifier),
            SkipFinalSnapshot=self.serialize(SkipFinalSnapshot),
            FinalDBSnapshotIdentifier=self.serialize(FinalDBSnapshotIdentifier),
            DeleteAutomatedBackups=self.serialize(DeleteAutomatedBackups),
        )
        _response = self.client.delete_db_instance(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DeleteDBInstanceResult(**_response)
        return cast("DBInstance", response.RDSDBInstance)

    def get(
        self, DBInstanceIdentifier: str, *, Filters: List[Filter] | None = None
    ) -> Optional["DBInstance"]:
        """
        Describes provisioned RDS instances. This API supports pagination.

        Args:
            DBInstanceIdentifier: The user-supplied instance identifier or the Amazon Resource Name (ARN) of the DB instance. If
                this parameter is specified, information from only the specific DB instance is returned. This parameter isn't case-
                sensitive.

        Keyword Args:
            Filters: A filter that specifies one or more DB instances to describe.
        """
        args: Dict[str, Any] = dict(
            DBInstanceIdentifier=self.serialize(DBInstanceIdentifier),
            Filters=self.serialize(Filters),
        )
        _response = self.client.describe_db_instances(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DBInstanceMessage(**_response)

        if response and response.DBInstances:
            self.sessionize(response.DBInstances[0])
            return response.DBInstances[0]
        return None

    def list(
        self,
        *,
        DBInstanceIdentifier: str | None = None,
        Filters: List[Filter] | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Describes provisioned RDS instances. This API supports pagination.

        Keyword Args:
            DBInstanceIdentifier: The user-supplied instance identifier or the Amazon Resource Name (ARN) of the DB instance. If
                this parameter is specified, information from only the specific DB instance is returned. This parameter isn't case-
                sensitive.
            Filters: A filter that specifies one or more DB instances to describe.
        """
        paginator = self.client.get_paginator("describe_db_instances")
        args: Dict[str, Any] = dict(
            DBInstanceIdentifier=self.serialize(DBInstanceIdentifier),
            Filters=self.serialize(Filters),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DBInstanceMessage(**_response)
            if response.DBInstances:
                results.extend(response.DBInstances)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results


class RDSDBSubnetGroupManager(Boto3ModelManager):
    service_name: str = "rds"

    def create(
        self, model: "RDSDBSubnetGroup", SubnetIds: List[str]
    ) -> "RDSDBSubnetGroup":
        """
        Creates a new DB subnet group. DB subnet groups must contain at least one subnet
        in at least two AZs in the Amazon Web Services Region.

        Args:
            model: The :py:class:`DBSubnetGroup` to create.
            SubnetIds: The EC2 Subnet IDs for the DB subnet group.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            DBSubnetGroupName=data.get("DBSubnetGroupName"),
            DBSubnetGroupDescription=data.get("DBSubnetGroupDescription"),
            SubnetIds=self.serialize(SubnetIds),
            Tags=data.get("TagList"),
        )
        _response = self.client.create_db_subnet_group(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateDBSubnetGroupResult(**_response)

        self.sessionize(response.DBSubnetGroup)
        return cast("RDSDBSubnetGroup", response.DBSubnetGroup)

    def update(
        self, model: "RDSDBSubnetGroup", SubnetIds: List[str]
    ) -> "RDSDBSubnetGroup":
        """
        Modifies an existing DB subnet group. DB subnet groups must contain at least one
        subnet in at least two AZs in the Amazon Web Services Region.

        Args:
            model: The :py:class:`DBSubnetGroup` to update.
            SubnetIds: The EC2 subnet IDs for the DB subnet group.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            DBSubnetGroupName=data.get("DBSubnetGroupName"),
            SubnetIds=self.serialize(SubnetIds),
            DBSubnetGroupDescription=data.get("DBSubnetGroupDescription"),
        )
        _response = self.client.modify_db_subnet_group(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = ModifyDBSubnetGroupResult(**_response)

        self.sessionize(response.DBSubnetGroup)
        return cast("RDSDBSubnetGroup", response.DBSubnetGroup)

    def delete(self, DBSubnetGroupName: str) -> None:
        """
        Deletes a DB subnet group.

        Args:
            DBSubnetGroupName: The name of the database subnet group to delete.
        """
        args: Dict[str, Any] = dict(DBSubnetGroupName=self.serialize(DBSubnetGroupName))
        self.client.delete_db_subnet_group(
            **{k: v for k, v in args.items() if v is not None}
        )

    def get(self, DBSubnetGroupName: str) -> Optional["RDSDBSubnetGroup"]:
        """
        Returns a list of DBSubnetGroup descriptions. If a DBSubnetGroupName is
        specified, the list will contain only the descriptions of the specified
        DBSubnetGroup.

        Args:
            DBSubnetGroupName: The name of the DB subnet group to return details for.
        """
        args: Dict[str, Any] = dict(DBSubnetGroupName=self.serialize(DBSubnetGroupName))
        _response = self.client.describe_db_subnet_groups(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DBSubnetGroupMessage(**_response)

        if response and response.DBSubnetGroups:
            self.sessionize(response.DBSubnetGroups[0])
            return response.DBSubnetGroups[0]
        return None

    def list(
        self,
        *,
        DBSubnetGroupName: str | None = None,
        Filters: List[Filter] | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Returns a list of DBSubnetGroup descriptions. If a DBSubnetGroupName is
        specified, the list will contain only the descriptions of the specified
        DBSubnetGroup.

        Keyword Args:
            DBSubnetGroupName: The name of the DB subnet group to return details for.
            Filters: This parameter isn't currently supported.
        """
        paginator = self.client.get_paginator("describe_db_subnet_groups")
        args: Dict[str, Any] = dict(
            DBSubnetGroupName=self.serialize(DBSubnetGroupName),
            Filters=self.serialize(Filters),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DBSubnetGroupMessage(**_response)
            if response.DBSubnetGroups:
                results.extend(response.DBSubnetGroups)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results


# ==============
# Service Models
# ==============


class RDSEndpoint(Boto3Model):
    """This data type represents the information you need to connect to an Amazon RDS DB instance. This data type is used as a
    response element in the following actions:

    * ``CreateDBInstance``
    * ``DescribeDBInstances``
    * ``DeleteDBInstance``

    For the data structure that represents Amazon Aurora DB cluster endpoints, see ``DBClusterEndpoint``.
    """

    Address: str | None = None
    """
    Specifies the DNS address of the DB instance.
    """
    Port: int | None = None
    """
    Specifies the port that the database engine is listening on.
    """
    HostedZoneId: str | None = None
    """
    Specifies the ID that Amazon Route 53 assigns when you create a hosted zone.
    """


class DBSecurityGroupMembership(Boto3Model):
    """This data type is used as a response element in the following actions:

    * ``ModifyDBInstance``
    * ``RebootDBInstance``
    * ``RestoreDBInstanceFromDBSnapshot``
    * ``RestoreDBInstanceToPointInTime``
    """

    DBSecurityGroupName: str | None = None
    """
    The name of the DB security group.
    """
    Status: str | None = None
    """
    The status of the DB security group.
    """


class VpcSecurityGroupMembership(Boto3Model):
    """
    This data type is used as a response element for queries on VPC security group
    membership.
    """

    VpcSecurityGroupId: str | None = None
    """
    The name of the VPC security group.
    """
    Status: str | None = None
    """
    The membership status of the VPC security group.
    """


class DBParameterGroupStatus(Boto3Model):
    """
    The status of the DB parameter group.

    This data type is used as a response element in the following actions:

    * ``CreateDBInstance``
    * ``CreateDBInstanceReadReplica``
    * ``DeleteDBInstance``
    * ``ModifyDBInstance``
    * ``RebootDBInstance``
    * ``RestoreDBInstanceFromDBSnapshot``
    """

    DBParameterGroupName: str | None = None
    """
    The name of the DB parameter group.
    """
    ParameterApplyStatus: str | None = None
    """
    The status of parameter updates.

    Valid values are:
    """


class RDSPendingCloudwatchLogsExports(Boto3Model):
    """
    A list of the log types whose configuration is still pending.

    In other words, these log types are in the process of being activated or
    deactivated.
    """

    LogTypesToEnable: List[str] | None = None
    """
    Log types that are in the process of being deactivated.

    After they are deactivated, these log types aren't exported to CloudWatch Logs.
    """
    LogTypesToDisable: List[str] | None = None
    """
    Log types that are in the process of being enabled.

    After they are enabled, these log types are exported to CloudWatch Logs.
    """


class ProcessorFeature(Boto3Model):
    """
    Contains the processor features of a DB instance class.

    To specify the number of CPU cores, use the ``coreCount`` feature name for the ``Name`` parameter. To specify the number
    of threads per core, use the ``threadsPerCore`` feature name for the ``Name`` parameter.

    You can set the processor features of the DB instance class for a DB instance when you call one of the following
    actions:

    * ``CreateDBInstance``
    * ``ModifyDBInstance``
    * ``RestoreDBInstanceFromDBSnapshot``
    * ``RestoreDBInstanceFromS3``
    * ``RestoreDBInstanceToPointInTime``

    You can view the valid processor values for a particular instance class by calling the
    ``DescribeOrderableDBInstanceOptions`` action and specifying the instance class for the ``DBInstanceClass`` parameter.

    In addition, you can use the following actions for DB instance class processor information:

    * ``DescribeDBInstances``
    * ``DescribeDBSnapshots``
    * ``DescribeValidDBInstanceModifications``

    If you call ``DescribeDBInstances``, ``ProcessorFeature`` returns non-null values only if the following conditions are
    met:

    * You are accessing an Oracle DB instance.
    * Your Oracle DB instance class supports configuring the number of CPU cores and threads per core.
    * The current number CPU cores and threads is set to a non-default value.

    For more information, see  `Configuring the processor for a DB instance class in RDS for
    Oracle <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.DBInstanceClass.html#USER_ConfigureProcessor>`_ in
    the *Amazon RDS User Guide.*
    """

    Name: str | None = None
    """
    The name of the processor feature.

    Valid names are ``coreCount`` and ``threadsPerCore``.
    """
    Value: str | None = None
    """
    The value of a processor feature.
    """


class RDSPendingModifiedValues(Boto3Model):
    """
    This data type is used as a response element in the ``ModifyDBInstance`` operation
    and contains changes that will be applied during the next maintenance window.
    """

    DBInstanceClass: str | None = None
    """
    The name of the compute and memory capacity class for the DB instance.
    """
    AllocatedStorage: int | None = None
    """
    The allocated storage size for the DB instance specified in gibibytes (GiB).
    """
    MasterUserPassword: str | None = None
    """
    The master credentials for the DB instance.
    """
    Port: int | None = None
    """
    The port for the DB instance.
    """
    BackupRetentionPeriod: int | None = None
    """
    The number of days for which automated backups are retained.
    """
    MultiAZ: bool | None = None
    """
    Indicates whether the Single-AZ DB instance will change to a Multi-AZ deployment.
    """
    EngineVersion: str | None = None
    """
    The database engine version.
    """
    LicenseModel: str | None = None
    """
    The license model for the DB instance.
    """
    Iops: int | None = None
    """
    The Provisioned IOPS value for the DB instance.
    """
    DBInstanceIdentifier: str | None = None
    """
    The database identifier for the DB instance.
    """
    StorageType: str | None = None
    """
    The storage type of the DB instance.
    """
    CACertificateIdentifier: str | None = None
    """
    The identifier of the CA certificate for the DB instance.
    """
    DBSubnetGroupName: str | None = None
    """
    The DB subnet group for the DB instance.
    """
    PendingCloudwatchLogsExports: RDSPendingCloudwatchLogsExports | None = None
    """
    A list of the log types whose configuration is still pending.

    In other words, these log types are in the process of being activated or
    deactivated.
    """
    ProcessorFeatures: List["ProcessorFeature"] | None = None
    """
    The number of CPU cores and the number of threads per core for the DB instance class
    of the DB instance.
    """
    IAMDatabaseAuthenticationEnabled: bool | None = None
    """
    Indicates whether mapping of Amazon Web Services Identity and Access Management
    (IAM) accounts to database accounts is enabled.
    """
    AutomationMode: Literal["full", "all-paused"] | None = None
    """
The automation mode of the RDS Custom DB instance: ``full`` or ``all-paused``. If ``full``, the DB instance automates
monitoring and instance recovery. If ``all-paused``, the instance pauses automation for the duration set by ``--resume-
full-automation-mode-minutes``.
    """
    ResumeFullAutomationModeTime: datetime | None = None
    """
    The number of minutes to pause the automation.

    When the time period ends, RDS Custom resumes full automation. The minimum value is
    60 (default). The maximum value is 1,440.
    """
    StorageThroughput: int | None = None
    """
    The storage throughput of the DB instance.
    """
    Engine: str | None = None
    """
    The database engine of the DB instance.
    """
    DedicatedLogVolume: bool | None = None
    """
    Indicates whether the DB instance has a dedicated log volume (DLV) enabled.>
    """
    MultiTenant: bool | None = None
    """
    Indicates whether the DB instance will change to the multi-tenant configuration
    (TRUE) or the single-tenant configuration (FALSE).
    """


class OptionGroupMembership(Boto3Model):
    """
    Provides information on the option groups the DB instance is a member of.
    """

    OptionGroupName: str | None = None
    """
    The name of the option group that the instance belongs to.
    """
    Status: str | None = None
    """
    The status of the DB instance's option group membership.

    Valid values are: ``in-sync``, ``pending-apply``, ``pending-
    removal``, ``pending-maintenance-apply``, ``pending-maintenance-removal``, ``applying``, ``removing``, and ``failed``.
    """


class DBInstanceStatusInfo(Boto3Model):
    """
    Provides a list of status information for a DB instance.
    """

    StatusType: str | None = None
    """
    This value is currently "read replication.".
    """
    Normal: bool | None = None
    """
    Indicates whether the instance is operating normally (TRUE) or is in an error state
    (FALSE).
    """
    Status: str | None = None
    """
    The status of the DB instance.

    For a StatusType of read replica, the values can be replicating, replication stop
    point set, replication stop point reached, error, stopped, or terminated.
    """
    Message: str | None = None
    """
    Details of the error if there is an error for the instance.

    If the instance isn't in an error state, this value is blank.
    """


class DomainMembership(Boto3Model):
    """
    An Active Directory Domain membership record associated with the DB instance or
    cluster.
    """

    Domain: str | None = None
    """
    The identifier of the Active Directory Domain.
    """
    Status: str | None = None
    """
    The status of the Active Directory Domain membership for the DB instance or cluster.

    Values include ``joined``,
    ``pending-join``, ``failed``, and so on.
    """
    FQDN: str | None = None
    """
    The fully qualified domain name (FQDN) of the Active Directory Domain.
    """
    IAMRoleName: str | None = None
    """
    The name of the IAM role used when making API calls to the Directory Service.
    """
    OU: str | None = None
    """
    The Active Directory organizational unit for the DB instance or cluster.
    """
    AuthSecretArn: str | None = None
    """
    The ARN for the Secrets Manager secret with the credentials for the user that's a
    member of the domain.
    """
    DnsIps: List[str] | None = None
    """
    The IPv4 DNS IP addresses of the primary and secondary Active Directory domain
    controllers.
    """


class DBInstanceRole(Boto3Model):
    """
    Information about an Amazon Web Services Identity and Access Management (IAM) role
    that is associated with a DB instance.
    """

    RoleArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the IAM role that is associated with the DB
    instance.
    """
    FeatureName: str | None = None
    """
    The name of the feature associated with the Amazon Web Services Identity and Access
    Management (IAM) role.

    For
    information about supported feature names, see ``DBEngineVersion``.
    """
    Status: str | None = None
    """
    Information about the state of association between the IAM role and the DB instance.

    The Status property returns one of
    the following values:
    """


class DBInstanceAutomatedBackupsReplication(Boto3Model):
    """
    Automated backups of a DB instance replicated to another Amazon Web Services Region.

    They consist of system backups, transaction logs, and database instance properties.
    """

    DBInstanceAutomatedBackupsArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the replicated automated backups.
    """


class RDSMasterUserSecret(Boto3Model):
    """
    Contains the secret managed by RDS in Amazon Web Services Secrets Manager for the
    master user password.

    For more information, see
    `Password management with Amazon Web Services Secrets Manager <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/rds-secrets-manager.html>`_
    in the *Amazon RDS User Guide*    and
    `Password management with Amazon Web Services Secrets Manager <https://docs.aws.amazon.com/AmazonRDS/latest/AuroraUserGuide/rds-secrets-manager.html>`_
    in the *Amazon Aurora
    User Guide.*
    """

    SecretArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the secret.
    """
    SecretStatus: str | None = None
    """
    The status of the secret.
    """
    KmsKeyId: str | None = None
    """
    The Amazon Web Services KMS key identifier that is used to encrypt the secret.
    """


class RDSCertificateDetails(Boto3Model):
    """
    The details of the DB instance’s server certificate.

    For more information, see `Using SSL/TLS to encrypt a connection to a DB
    instance <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/UsingWithRDS.SSL.html>`_ in the *Amazon RDS User Guide*
    and  `Using SSL/TLS to encrypt a connection to a DB
    cluster <https://docs.aws.amazon.com/AmazonRDS/latest/AuroraUserGuide/UsingWithRDS.SSL.html>`_ in the *Amazon Aurora User
    Guide*.
    """

    CAIdentifier: str | None = None
    """
    The CA identifier of the CA certificate used for the DB instance's server
    certificate.
    """
    ValidTill: datetime | None = None
    """
    The expiration date of the DB instance’s server certificate.
    """


class DBInstance(TagsDictMixin, PrimaryBoto3Model):
    """
    Contains the details of an Amazon RDS DB instance.

    This data type is used as a response element in the operations ``CreateDBInstance``, ``CreateDBInstanceReadReplica``,
    ``DeleteDBInstance``, ``DescribeDBInstances``, ``ModifyDBInstance``, ``PromoteReadReplica``, ``RebootDBInstance``,
    ``RestoreDBInstanceFromDBSnapshot``, ``RestoreDBInstanceFromS3``, ``RestoreDBInstanceToPointInTime``,
    ``StartDBInstance``, and ``StopDBInstance``.
    """

    tag_class: ClassVar[Type] = Tag
    manager_class: ClassVar[Type[Boto3ModelManager]] = DBInstanceManager

    DBInstanceIdentifier: str
    """
    The user-supplied database identifier.

    This identifier is the unique key that identifies a DB instance.
    """
    DBInstanceClass: str
    """
    The name of the compute and memory capacity class of the DB instance.
    """
    MultiAZ: bool | None
    """
    Indicates whether the DB instance is a Multi-AZ deployment.

    This setting doesn't apply to RDS Custom DB instances.
    """
    Engine: str
    """
    The database engine used for this DB instance.
    """
    EngineVersion: str
    """
    The version of the database engine.
    """
    AutoMinorVersionUpgrade: bool | None = True
    """
    Indicates whether minor version patches are applied automatically.
    """
    PubliclyAccessible: bool | None
    """
    Indicates whether the DB instance is publicly accessible.
    """
    StorageEncrypted: bool | None = True
    """
    Indicates whether the DB instance is encrypted.
    """
    BackupRetentionPeriod: int | None = 1
    """
    The number of days for which automatic DB snapshots are retained.
    """
    DbInstancePort: int | None = None
    """
    The port that the DB instance listens on.

    If the DB instance is part of a DB cluster, this can be a different port than the DB
    cluster port.
    """
    IAMDatabaseAuthenticationEnabled: bool | None = None
    """
    Indicates whether mapping of Amazon Web Services Identity and Access Management
    (IAM) accounts to database accounts is enabled for the DB instance.
    """
    EnabledCloudwatchLogsExports: List[str] | None = None
    """
    A list of log types that this DB instance is configured to export to CloudWatch
    Logs.
    """
    Tags: List[Tag] = Field(default_factory=list, alias="TagList")
    """
    A list of tags.
    """
    CustomerOwnedIpEnabled: bool | None = None
    """
    Indicates whether a customer-owned IP address (CoIP) is enabled for an RDS on
    Outposts DB instance.
    """
    PerformanceInsightsEnabled: bool | None = None
    """
    Indicates whether Performance Insights is enabled for the DB instance.
    """
    DBSubnetGroup: Optional["RDSDBSubnetGroup"] = None
    """
    Information about the subnet group associated with the DB instance, including the
    name, description, and subnets in the subnet group.
    """
    DBInstanceStatus: str = Field(default=None, frozen=True)
    """
    The current state of this database.
    """
    AutomaticRestartTime: datetime = Field(default=None, frozen=True)
    """
    The time when a stopped DB instance is restarted automatically.
    """
    MasterUsername: str | None = None
    """
    The master username for the DB instance.
    """
    DBName: str | None = None
    """
    The initial database name that you provided (if required) when you created the DB
    instance.

    This name is returned for the life of your DB instance. For an RDS for Oracle CDB
    instance, the name identifies the PDB rather than the CDB.
    """
    Endpoint: RDSEndpoint = Field(default=None, frozen=True)
    """
    The connection endpoint for the DB instance.
    """
    AllocatedStorage: int | None = None
    """
    The amount of storage in gibibytes (GiB) allocated for the DB instance.
    """
    InstanceCreateTime: datetime = Field(default=None, frozen=True)
    """
    The date and time when the DB instance was created.
    """
    PreferredBackupWindow: str | None = None
    """
    The daily time range during which automated backups are created if automated backups
    are enabled, as determined by the ``BackupRetentionPeriod``.
    """
    DBSecurityGroups: List["DBSecurityGroupMembership"] | None = None
    """
    A list of DB security group elements containing ``DBSecurityGroup.Name`` and
    ``DBSecurityGroup.Status`` subelements.
    """
    VpcSecurityGroups: List["VpcSecurityGroupMembership"] = Field(
        default_factory=list, frozen=True
    )
    """
    The list of Amazon EC2 VPC security groups that the DB instance belongs to.
    """
    DBParameterGroups: List["DBParameterGroupStatus"] = Field(
        default_factory=list, frozen=True
    )
    """
    The list of DB parameter groups applied to this DB instance.
    """
    AvailabilityZone: str | None = None
    """
    The name of the Availability Zone where the DB instance is located.
    """
    PreferredMaintenanceWindow: str | None = None
    """
    The weekly time range during which system maintenance can occur, in Universal
    Coordinated Time (UTC).
    """
    PendingModifiedValues: RDSPendingModifiedValues = Field(default=None, frozen=True)
    """
    Information about pending changes to the DB instance.

    This information is returned only when there are pending changes. Specific changes
    are identified by subelements.
    """
    LatestRestorableTime: datetime = Field(default=None, frozen=True)
    """
    The latest time to which a database in this DB instance can be restored with point-
    in-time restore.
    """
    ReadReplicaSourceDBInstanceIdentifier: str = Field(default=None, frozen=True)
    """
    The identifier of the source DB instance if this DB instance is a read replica.
    """
    ReadReplicaDBInstanceIdentifiers: List[str] = Field(
        default_factory=list, frozen=True
    )
    """
    The identifiers of the read replicas associated with this DB instance.
    """
    ReadReplicaDBClusterIdentifiers: List[str] = Field(
        default_factory=list, frozen=True
    )
    """
    The identifiers of Aurora DB clusters to which the RDS DB instance is replicated as
    a read replica.

    For example, when you create an Aurora read replica of an RDS for MySQL DB instance,
    the Aurora MySQL DB cluster for the Aurora read replica is shown. This output
    doesn't contain information about cross-Region Aurora read replicas.
    """
    ReplicaMode: Literal["open-read-only", "mounted"] | None = None
    """
    The open mode of a Db2 or an Oracle read replica.

    The default is ``open-read-only``. For more information, see `Working
    with replicas for Amazon RDS for Db2 <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/db2-replication.html>`_ and
    `Working with read replicas for Amazon RDS for Oracle <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/oracle-
    read-replicas.html>`_ in the *Amazon RDS User Guide*.
    """
    LicenseModel: str | None = None
    """
    The license model information for this DB instance.

    This setting doesn't apply to Amazon Aurora or RDS Custom DB instances.
    """
    Iops: int | None = None
    """
    The Provisioned IOPS (I/O operations per second) value for the DB instance.
    """
    OptionGroupMemberships: List["OptionGroupMembership"] = Field(
        default_factory=list, frozen=True
    )
    """
    The list of option group memberships for this DB instance.
    """
    CharacterSetName: str | None = None
    """
    If present, specifies the name of the character set that this instance is associated
    with.
    """
    NcharCharacterSetName: str | None = None
    """
    The name of the NCHAR character set for the Oracle DB instance.

    This character set specifies the Unicode encoding for data stored in table columns
    of type NCHAR, NCLOB, or NVARCHAR2.
    """
    SecondaryAvailabilityZone: str = Field(default=None, frozen=True)
    """
    If present, specifies the name of the secondary Availability Zone for a DB instance
    with multi-AZ support.
    """
    StatusInfos: List["DBInstanceStatusInfo"] = Field(default_factory=list, frozen=True)
    """
    The status of a read replica.

    If the DB instance isn't a read replica, the value is blank.
    """
    StorageType: str | None = None
    """
    The storage type associated with the DB instance.
    """
    TdeCredentialArn: str | None = None
    """
    The ARN from the key store with which the instance is associated for TDE encryption.
    """
    DBClusterIdentifier: str | None = None
    """
    If the DB instance is a member of a DB cluster, indicates the name of the DB cluster
    that the DB instance is a member of.
    """
    KmsKeyId: str | None = None
    """
    If ``StorageEncrypted`` is enabled, the Amazon Web Services KMS key identifier for
    the encrypted DB instance.
    """
    DbiResourceId: str = Field(default=None, frozen=True)
    """
    The Amazon Web Services Region-unique, immutable identifier for the DB instance.

    This identifier is found in Amazon Web Services CloudTrail log entries whenever the
    Amazon Web Services KMS key for the DB instance is accessed.
    """
    CACertificateIdentifier: str | None = None
    """
    The identifier of the CA certificate for this DB instance.
    """
    DomainMemberships: List["DomainMembership"] = Field(
        default_factory=list, frozen=True
    )
    """
    The Active Directory Domain membership records associated with the DB instance.
    """
    CopyTagsToSnapshot: bool | None = None
    """
    Indicates whether tags are copied from the DB instance to snapshots of the DB
    instance.
    """
    MonitoringInterval: int | None = None
    """
    The interval, in seconds, between points when Enhanced Monitoring metrics are
    collected for the DB instance.
    """
    EnhancedMonitoringResourceArn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) of the Amazon CloudWatch Logs log stream that
    receives the Enhanced Monitoring metrics data for the DB instance.
    """
    MonitoringRoleArn: str | None = None
    """
    The ARN for the IAM role that permits RDS to send Enhanced Monitoring metrics to
    Amazon CloudWatch Logs.
    """
    PromotionTier: int | None = None
    """
    The order of priority in which an Aurora Replica is promoted to the primary instance
    after a failure of the existing primary instance.

    For more information, see  `Fault Tolerance for an Aurora DB Cluster <https://docs.aws.amazon.com/Amaz
    onRDS/latest/AuroraUserGuide/Concepts.AuroraHighAvailability.html#Aurora.Managing.FaultTolerance>`_ in the *Amazon Aurora
    User Guide*.
    """
    DBInstanceArn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) for the DB instance.
    """
    Timezone: str | None = None
    """
    The time zone of the DB instance.

    In most cases, the ``Timezone`` element is empty. ``Timezone`` content appears only
    for RDS for Db2 and RDS for SQL Server DB instances that were created with a time zone specified.
    """
    DatabaseInsightsMode: Literal["standard", "advanced"] | None = None
    """
    The mode of Database Insights that is enabled for the instance.
    """
    PerformanceInsightsKMSKeyId: str | None = None
    """
    The Amazon Web Services KMS key identifier for encryption of Performance Insights
    data.
    """
    PerformanceInsightsRetentionPeriod: int | None = None
    """
    The number of days to retain Performance Insights data.
    """
    ProcessorFeatures: List["ProcessorFeature"] | None = None
    """
    The number of CPU cores and the number of threads per core for the DB instance class
    of the DB instance.
    """
    DeletionProtection: bool | None = None
    """
    Indicates whether the DB instance has deletion protection enabled.

    The database can't be deleted when deletion protection is enabled. For more
    information, see
    `Deleting a DB Instance <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/USER_DeleteInstance.html>`_.
    """
    AssociatedRoles: List["DBInstanceRole"] = Field(default_factory=list, frozen=True)
    """
    The Amazon Web Services Identity and Access Management (IAM) roles associated with
    the DB instance.
    """
    ListenerEndpoint: RDSEndpoint = Field(default=None, frozen=True)
    """
    The listener connection endpoint for SQL Server Always On.
    """
    MaxAllocatedStorage: int | None = None
    """
    The upper limit in gibibytes (GiB) to which Amazon RDS can automatically scale the
    storage of the DB instance.
    """
    DBInstanceAutomatedBackupsReplications: List[
        "DBInstanceAutomatedBackupsReplication"
    ] = Field(default_factory=list, frozen=True)
    """
    The list of replicated automated backups associated with the DB instance.
    """
    AwsBackupRecoveryPointArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the recovery point in Amazon Web Services Backup.
    """
    ActivityStreamStatus: Literal["stopped", "starting", "started", "stopping"] = Field(
        default=None, frozen=True
    )
    """
    The status of the database activity stream.
    """
    ActivityStreamKmsKeyId: str = Field(default=None, frozen=True)
    """
    The Amazon Web Services KMS key identifier used for encrypting messages in the
    database activity stream.

    The Amazon Web Services KMS key identifier is the key ARN, key ID, alias ARN, or
    alias name for the KMS key.
    """
    ActivityStreamKinesisStreamName: str = Field(default=None, frozen=True)
    """
    The name of the Amazon Kinesis data stream used for the database activity stream.
    """
    ActivityStreamMode: Literal["sync", "async"] = Field(default=None, frozen=True)
    """
    The mode of the database activity stream.

    Database events such as a change or access generate an activity stream event. RDS
    for Oracle always handles these events asynchronously.
    """
    ActivityStreamEngineNativeAuditFieldsIncluded: bool = Field(
        default=None, frozen=True
    )
    """
    Indicates whether engine-native audit fields are included in the database activity
    stream.
    """
    AutomationMode: Literal["full", "all-paused"] | None = None
    """
The automation mode of the RDS Custom DB instance: ``full`` or ``all paused``. If ``full``, the DB instance automates
monitoring and instance recovery. If ``all paused``, the instance pauses automation for the duration set by ``--resume-
full-automation-mode-minutes``.
    """
    ResumeFullAutomationModeTime: datetime = Field(default=None, frozen=True)
    """
    The number of minutes to pause the automation.

    When the time period ends, RDS Custom resumes full automation. The minimum value is
    60 (default). The maximum value is 1,440.
    """
    CustomIamInstanceProfile: str | None = None
    """
    The instance profile associated with the underlying Amazon EC2 instance of an RDS
    Custom DB instance.

    The instance
    profile must meet the following requirements:
    """
    BackupTarget: str | None = None
    """
The location where automated backups and manual snapshots are stored: Dedicated Local Zones, Amazon Web Services
Outposts or the Amazon Web Services Region.
    """
    NetworkType: str | None = None
    """
    The network type of the DB instance.
    """
    ActivityStreamPolicyStatus: Literal[
        "locked", "unlocked", "locking-policy", "unlocking-policy"
    ] = Field(default=None, frozen=True)
    """
    The status of the policy state of the activity stream.
    """
    StorageThroughput: int | None = None
    """
    The storage throughput for the DB instance.
    """
    DBSystemId: str | None = None
    """
    The Oracle system ID (Oracle SID) for a container database (CDB).

    The Oracle SID is also the name of the CDB. This setting is only valid for RDS
    Custom DB instances.
    """
    MasterUserSecret: RDSMasterUserSecret = Field(default=None, frozen=True)
    """
    The secret managed by RDS in Amazon Web Services Secrets Manager for the master user
    password.
    """
    CertificateDetails: RDSCertificateDetails = Field(default=None, frozen=True)
    """
    The details of the DB instance's server certificate.
    """
    ReadReplicaSourceDBClusterIdentifier: str = Field(default=None, frozen=True)
    """
    The identifier of the source DB cluster if this DB instance is a read replica.
    """
    PercentProgress: str = Field(default=None, frozen=True)
    """
    The progress of the storage optimization operation as a percentage.
    """
    DedicatedLogVolume: bool | None = None
    """
    Indicates whether the DB instance has a dedicated log volume (DLV) enabled.
    """
    IsStorageConfigUpgradeAvailable: bool = Field(default=None, frozen=True)
    """
    Indicates whether an upgrade is recommended for the storage file system
    configuration on the DB instance.

    To migrate to
    the preferred configuration, you can either create a blue/green deployment, or create a read replica from the DB
    instance. For more information, see `Upgrading the storage file system for a DB instance <https://docs.aws.amazon.com/Am
    azonRDS/latest/UserGuide/USER_PIOPS.StorageTypes.html#USER_PIOPS.UpgradeFileSystem>`_.
    """
    MultiTenant: bool | None = None
    """
    Specifies whether the DB instance is in the multi-tenant configuration (TRUE) or the
    single-tenant configuration (FALSE).
    """
    EngineLifecycleSupport: str | None = None
    """
    The lifecycle type for the DB instance.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`DBInstanceArn` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.DBInstanceArn

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`DBInstanceArn`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.DBInstanceArn

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`DBInstanceIdentifier` attribute.

        Returns:
            The name of the model instance.
        """
        return self.DBInstanceIdentifier

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`DBInstanceArn` attribute.
        """
        return hash(self.DBInstanceArn)

    @cached_property
    def subnet_group(self) -> Optional["RDSDBSubnetGroup"]:
        """
        Return the subnet group that this DB instance is in, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "DBSubnetGroupName": self.DBSubnetGroup.DBSubnetGroupName,
                }
            )
        except AttributeError:
            return None
        return RDSDBSubnetGroup.objects.using(self.session).get(**pk)  # type: ignore[arg-type]

    @cached_property
    def security_groups(self) -> List["SecurityGroup"] | None:
        """
        Return the security groups that this DB instance is in, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "GroupIds": [
                        g.VpcSecurityGroupId
                        for g in self.VpcSecurityGroups
                        if g.VpcSecurityGroupId
                    ],
                }
            )
        except AttributeError:
            return []
        return SecurityGroup.objects.using(self.session).list(**pk)  # type: ignore[arg-type]

    @cached_property
    def vpc(self) -> Optional["Vpc"]:
        """
        Return the VPC that this DB instance is in, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "VpcId": self.DBSubnetGroup.VpcId,
                }
            )
        except AttributeError:
            return None
        return Vpc.objects.using(self.session).get(**pk)  # type: ignore[arg-type]


class RDSAvailabilityZone(Boto3Model):
    """
    Contains Availability Zone information.

    This data type is used as an element in the ``OrderableDBInstanceOption`` data type.
    """

    Name: str | None = None
    """
    The name of the Availability Zone.
    """


class Outpost(Boto3Model):
    """
    A data type that represents an Outpost.

    For more information about RDS on Outposts, see
    `Amazon RDS on Amazon Web Services Outposts <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/rds-on-outposts.html>`_
    in the *Amazon RDS User Guide.*
    """

    Arn: str | None = None
    """
    The Amazon Resource Name (ARN) of the Outpost.
    """


class RDSSubnet(Boto3Model):
    """
    This data type is used as a response element for the ``DescribeDBSubnetGroups``
    operation.
    """

    SubnetIdentifier: str | None = None
    """
    The identifier of the subnet.
    """
    SubnetAvailabilityZone: RDSAvailabilityZone | None = None
    """
    Contains Availability Zone information.
    """
    SubnetOutpost: Outpost | None = None
    """
    If the subnet is associated with an Outpost, this value specifies the Outpost.
    """
    SubnetStatus: str | None = None
    """
    The status of the subnet.
    """


class RDSDBSubnetGroup(PrimaryBoto3Model):
    """
    Contains the details of an Amazon RDS DB subnet group.

    This data type is used as a response element in the ``DescribeDBSubnetGroups`` action.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = RDSDBSubnetGroupManager

    DBSubnetGroupName: str
    """
    The name of the DB subnet group.
    """
    DBSubnetGroupDescription: str
    """
    Provides the description of the DB subnet group.
    """
    Subnets: List["RDSSubnet"]
    """
    Contains a list of ``Subnet`` elements.

    The list of subnets shown here might not reflect the current state of your VPC. For
    the most up-to-date information, we recommend checking your VPC configuration
    directly.
    """
    VpcId: str
    """
    Provides the VpcId of the DB subnet group.
    """
    SubnetGroupStatus: str | None = "Complete"
    """
    Provides the status of the DB subnet group.
    """
    DBSubnetGroupArn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) for the DB subnet group.
    """
    SupportedNetworkTypes: List[str] = Field(default_factory=list, frozen=True)
    """
    The network type of the DB subnet group.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`DBSubnetGroupArn` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.DBSubnetGroupArn

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`DBSubnetGroupArn` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.DBSubnetGroupArn

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`DBSubnetGroupName` attribute.

        Returns:
            The name of the model instance.
        """
        return self.DBSubnetGroupName

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`DBSubnetGroupArn` attribute.
        """
        return hash(self.DBSubnetGroupArn)


# =======================
# Request/Response Models
# =======================


class CreateDBInstanceResult(Boto3Model):
    RDSDBInstance: DBInstance = Field(default=None, alias="DBInstance")
    """
    Contains the details of an Amazon RDS DB instance.
    """


class CloudwatchLogsExportConfiguration(Boto3Model):
    """
    The configuration setting for the log types to be enabled for export to CloudWatch
    Logs for a specific DB instance or DB cluster.

    The ``EnableLogTypes`` and ``DisableLogTypes`` arrays determine which logs will be exported (or not exported) to
    CloudWatch Logs. The values within these arrays depend on the DB engine being used.

    For more information about exporting CloudWatch Logs for Amazon RDS DB instances, see `Publishing Database Logs to
    Amazon CloudWatch Logs <https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/USER_LogAccess.html#USER_LogAccess.Proced
    ural.UploadtoCloudWatch>`_  in the *Amazon RDS User Guide*.

    For more information about exporting CloudWatch Logs for Amazon Aurora DB clusters, see `Publishing Database Logs to
    Amazon CloudWatch Logs <https://docs.aws.amazon.com/AmazonRDS/latest/AuroraUserGuide/USER_LogAccess.html#USER_LogAccess.
    Procedural.UploadtoCloudWatch>`_ in the *Amazon Aurora User Guide*.
    """

    EnableLogTypes: List[str] | None = None
    """
    The list of log types to enable.
    """
    DisableLogTypes: List[str] | None = None
    """
    The list of log types to disable.
    """


class ModifyDBInstanceResult(Boto3Model):
    RDSDBInstance: DBInstance = Field(default=None, alias="DBInstance")
    """
    Contains the details of an Amazon RDS DB instance.
    """


class DeleteDBInstanceResult(Boto3Model):
    RDSDBInstance: DBInstance = Field(default=None, alias="DBInstance")
    """
    Contains the details of an Amazon RDS DB instance.
    """


class DBInstanceMessage(Boto3Model):
    """
    Contains the result of a successful invocation of the ``DescribeDBInstances``
    action.
    """

    Marker: str | None = None
    """
    An optional pagination token provided by a previous request.

    If this parameter is specified, the response includes only
    records beyond the marker, up to the value specified by ``MaxRecords`` .
    """
    DBInstances: List["DBInstance"] | None = None
    """
    A list of ``DBInstance`` instances.
    """


class CreateDBSubnetGroupResult(Boto3Model):
    DBSubnetGroup: RDSDBSubnetGroup | None = None
    """
    Contains the details of an Amazon RDS DB subnet group.
    """


class ModifyDBSubnetGroupResult(Boto3Model):
    DBSubnetGroup: RDSDBSubnetGroup | None = None
    """
    Contains the details of an Amazon RDS DB subnet group.
    """


class DBSubnetGroupMessage(Boto3Model):
    """
    Contains the result of a successful invocation of the ``DescribeDBSubnetGroups``
    action.
    """

    Marker: str | None = None
    """
    An optional pagination token provided by a previous request.

    If this parameter is specified, the response includes only
    records beyond the marker, up to the value specified by ``MaxRecords``.
    """
    DBSubnetGroups: List["RDSDBSubnetGroup"] | None = None
    """
    A list of ``DBSubnetGroup`` instances.
    """



# This file is automatically generated by botocraft.  Do not edit directly.
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
from botocraft.mixins.tags import TagsDictMixin
from botocraft.services.common import Tag
from .abstract import (
    Boto3Model,
    ReadonlyBoto3Model,
    PrimaryBoto3Model,
    ReadonlyPrimaryBoto3Model,
    Boto3ModelManager,
    ReadonlyBoto3ModelManager,
)
from datetime import datetime
from typing import ClassVar, Type, Optional, Literal, Dict, List, Union, Any, cast
from pydantic import Field

# ===============
# Managers
# ===============


class AssumeRoleManager(Boto3ModelManager):
    service_name: str = "sts"

    def assume(
        self,
        RoleArn: str,
        RoleSessionName: str,
        *,
        PolicyArns: List["STSPolicyDescriptorType"] | None = None,
        Policy: str | None = None,
        DurationSeconds: int | None = None,
        Tags: List[Tag] | None = None,
        TransitiveTagKeys: List[str] | None = None,
        ExternalId: str | None = None,
        SerialNumber: str | None = None,
        TokenCode: str | None = None,
        SourceIdentity: str | None = None,
        ProvidedContexts: List["STSProvidedContext"] | None = None,
    ) -> "AssumeRole":
        """
        Returns a set of temporary security credentials that you can use to access Amazon Web Services resources. These
        temporary credentials consist of an access key ID, a secret access key, and a security token. Typically, you use
        ``AssumeRole`` within your account or for cross-account access. For a comparison of ``AssumeRole`` with other API
        operations that produce temporary credentials, see `Requesting Temporary Security
        Credentials <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_temp_request.html>`_ and `Compare STS
        credentials <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_sts-comparison.html>`_ in the *IAM User
        Guide*.

        Args:
            RoleArn: The Amazon Resource Name (ARN) of the role to assume.
            RoleSessionName: An identifier for the assumed role session.

        Keyword Args:
            PolicyArns: The Amazon Resource Names (ARNs) of the IAM managed policies that you want to use as managed session
                policies. The policies must exist in the same account as the role.
            Policy: An IAM policy in JSON format that you want to use as an inline session policy.
            DurationSeconds: The duration, in seconds, of the role session. The value specified can range from 900 seconds (15
                minutes) up to the maximum session duration set for the role. The maximum session duration setting can have a value
                from 1 hour to 12 hours. If you specify a value higher than this setting or the administrator setting (whichever is
                lower), the operation fails. For example, if you specify a session duration of 12 hours, but your administrator set
                the maximum session duration to 6 hours, your operation fails.
            Tags: A list of session tags that you want to pass. Each session tag consists of a key name and an associated value.
                For more information about session tags, see `Tagging Amazon Web Services STS Sessions
                <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_session-tags.html>`_ in the *IAM User Guide*.
            TransitiveTagKeys: A list of keys for session tags that you want to set as transitive. If you set a tag key as
                transitive, the corresponding key and value passes to subsequent sessions in a role chain. For more information, see
                `Chaining Roles with Session Tags <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_session-tags.html#id_session-
                tags_role-chaining>`_ in the *IAM User Guide*.
            ExternalId: A unique identifier that might be required when you assume a role in another account. If the
                administrator of the account to which the role belongs provided you with an external ID, then provide that value in
                the ``ExternalId`` parameter. This value can be any string, such as a passphrase or account number. A cross-account
                role is usually set up to trust everyone in an account. Therefore, the administrator of the trusting account might
                send an external ID to the administrator of the trusted account. That way, only someone with the ID can assume the
                role, rather than everyone in the account. For more information about the external ID, see `How to Use an External
                ID When Granting Access to Your Amazon Web Services Resources to a Third Party
                <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_create_for- user_externalid.html>`_ in the *IAM User
                Guide*.
            SerialNumber: The identification number of the MFA device that is associated with the user who is making the
                ``AssumeRole`` call. Specify this value if the trust policy of the role being assumed includes a condition that
                requires MFA authentication. The value is either the serial number for a hardware device (such as ``GAHT12345678``)
                or an Amazon Resource Name (ARN) for a virtual device (such as ``arn:aws:iam::123456789012:mfa/user``).
            TokenCode: The value provided by the MFA device, if the trust policy of the role being assumed requires MFA. (In
                other words, if the policy includes a condition that tests for MFA). If the role being assumed requires MFA and if
                the ``TokenCode`` value is missing or expired, the ``AssumeRole`` call returns an "access denied" error.
            SourceIdentity: The source identity specified by the principal that is calling the ``AssumeRole`` operation. The
                source identity value persists across `chained role
                <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles.html#iam-term-role-chaining>`_ sessions.
            ProvidedContexts: A list of previously acquired trusted context assertions in the format of a JSON array. The
                trusted context assertion is signed and encrypted by Amazon Web Services STS.

        """
        args: Dict[str, Any] = dict(
            RoleArn=self.serialize(RoleArn),
            RoleSessionName=self.serialize(RoleSessionName),
            PolicyArns=self.serialize(PolicyArns),
            Policy=self.serialize(Policy),
            DurationSeconds=self.serialize(DurationSeconds),
            Tags=self.serialize(Tags),
            TransitiveTagKeys=self.serialize(TransitiveTagKeys),
            ExternalId=self.serialize(ExternalId),
            SerialNumber=self.serialize(SerialNumber),
            TokenCode=self.serialize(TokenCode),
            SourceIdentity=self.serialize(SourceIdentity),
            ProvidedContexts=self.serialize(ProvidedContexts),
        )
        _response = self.client.assume_role(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = AssumeRole(**_response)

        results: "AssumeRole" = None
        if response is not None:
            results = response

        self.sessionize(results)
        return cast("AssumeRole", results)

    def assume_with_saml(
        self,
        RoleArn: str,
        PrincipalArn: str,
        SAMLAssertion: str,
        *,
        PolicyArns: List["STSPolicyDescriptorType"] | None = None,
        Policy: str | None = None,
        DurationSeconds: int | None = None,
    ) -> Optional["AssumeRole"]:
        """
        Returns a set of temporary security credentials for users who have been authenticated via a SAML authentication
        response. This operation provides a mechanism for tying an enterprise identity store or directory to role-based Amazon
        Web Services access without user-specific credentials or configuration. For a comparison of ``AssumeRoleWithSAML`` with
        the other API operations that produce temporary credentials, see `Requesting Temporary Security
        Credentials <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_temp_request.html>`_ and `Compare STS
        credentials <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_sts-comparison.html>`_ in the *IAM User
        Guide*.

        Args:
            RoleArn: The Amazon Resource Name (ARN) of the role that the caller is assuming.
            PrincipalArn: The Amazon Resource Name (ARN) of the SAML provider in IAM that describes the IdP.
            SAMLAssertion: The base64 encoded SAML authentication response provided by the IdP.

        Keyword Args:
            PolicyArns: The Amazon Resource Names (ARNs) of the IAM managed policies that you want to use as managed session
                policies. The policies must exist in the same account as the role.
            Policy: An IAM policy in JSON format that you want to use as an inline session policy.
            DurationSeconds: The duration, in seconds, of the role session. Your role session lasts for the duration that you
                specify for the ``DurationSeconds`` parameter, or until the time specified in the SAML authentication response's
                ``SessionNotOnOrAfter`` value, whichever is shorter. You can provide a ``DurationSeconds`` value from 900 seconds
                (15 minutes) up to the maximum session duration setting for the role. This setting can have a value from 1 hour to
                12 hours. If you specify a value higher than this setting, the operation fails. For example, if you specify a
                session duration of 12 hours, but your administrator set the maximum session duration to 6 hours, your operation
                fails. To learn how to view the maximum value for your role, see `View the Maximum Session Duration Setting for a
                Role <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_use.html#id_roles_use_view-role-max-session>`_ in
                the *IAM User Guide*.

        """
        args: Dict[str, Any] = dict(
            RoleArn=self.serialize(RoleArn),
            PrincipalArn=self.serialize(PrincipalArn),
            SAMLAssertion=self.serialize(SAMLAssertion),
            PolicyArns=self.serialize(PolicyArns),
            Policy=self.serialize(Policy),
            DurationSeconds=self.serialize(DurationSeconds),
        )
        _response = self.client.assume_role_with_saml(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = AssumeRoleWithSAMLResponse(**_response)

        results: "AssumeRole" | None = None
        if response is not None:
            results = response

        self.sessionize(results)
        return cast("AssumeRole | None", results)

    def assume_with_web_identity(
        self,
        RoleArn: str,
        RoleSessionName: str,
        WebIdentityToken: str,
        *,
        ProviderId: str | None = None,
        PolicyArns: List["STSPolicyDescriptorType"] | None = None,
        Policy: str | None = None,
        DurationSeconds: int | None = None,
    ) -> Optional["AssumeRole"]:
        """
        Returns a set of temporary security credentials for users who have been authenticated in a mobile or web application
        with a web identity provider. Example providers include the OAuth 2.0 providers Login with Amazon and Facebook, or any
        OpenID Connect-compatible identity provider such as Google or `Amazon Cognito federated
        identities <https://docs.aws.amazon.com/cognito/latest/developerguide/cognito-identity.html>`_.

        Args:
            RoleArn: The Amazon Resource Name (ARN) of the role that the caller is assuming.
            RoleSessionName: An identifier for the assumed role session. Typically, you pass the name or identifier that is
                associated with the user who is using your application. That way, the temporary security credentials that your
                application will use are associated with that user. This session name is included as part of the ARN and assumed
                role ID in the ``AssumedRoleUser`` response element.
            WebIdentityToken: The OAuth 2.0 access token or OpenID Connect ID token that is provided by the identity provider.
                Your application must get this token by authenticating the user who is using your application with a web identity
                provider before the application makes an ``AssumeRoleWithWebIdentity`` call. Timestamps in the token must be
                formatted as either an integer or a long integer. Tokens must be signed using either RSA keys (RS256, RS384, or
                RS512) or ECDSA keys (ES256, ES384, or ES512).

        Keyword Args:
            ProviderId: The fully qualified host component of the domain name of the OAuth 2.0 identity provider. Do not specify
                this value for an OpenID Connect identity provider.
            PolicyArns: The Amazon Resource Names (ARNs) of the IAM managed policies that you want to use as managed session
                policies. The policies must exist in the same account as the role.
            Policy: An IAM policy in JSON format that you want to use as an inline session policy.
            DurationSeconds: The duration, in seconds, of the role session. The value can range from 900 seconds (15 minutes) up
                to the maximum session duration setting for the role. This setting can have a value from 1 hour to 12 hours. If you
                specify a value higher than this setting, the operation fails. For example, if you specify a session duration of 12
                hours, but your administrator set the maximum session duration to 6 hours, your operation fails. To learn how to
                view the maximum value for your role, see `View the Maximum Session Duration Setting for a Role
                <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_use.html#id_roles_use_view-role-max-session>`_ in the
                *IAM User Guide*.

        """
        args: Dict[str, Any] = dict(
            RoleArn=self.serialize(RoleArn),
            RoleSessionName=self.serialize(RoleSessionName),
            WebIdentityToken=self.serialize(WebIdentityToken),
            ProviderId=self.serialize(ProviderId),
            PolicyArns=self.serialize(PolicyArns),
            Policy=self.serialize(Policy),
            DurationSeconds=self.serialize(DurationSeconds),
        )
        _response = self.client.assume_role_with_web_identity(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = AssumeRoleWithWebIdentityResponse(**_response)

        results: "AssumeRole" | None = None
        if response is not None:
            results = response

        self.sessionize(results)
        return cast("AssumeRole | None", results)


class CallerIdentityManager(Boto3ModelManager):
    service_name: str = "sts"

    def get(
        self,
    ) -> Optional["CallerIdentity"]:
        """
        Returns details about the IAM user or role whose credentials are used to call
        the operation.
        """
        args: Dict[str, Any] = dict()
        _response = self.client.get_caller_identity(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CallerIdentity(**_response)

        if response:
            self.sessionize(response)
            return response
        return None


# ==============
# Service Models
# ==============


class STSCredentials(Boto3Model):
    """
    Amazon Web Services credentials for API authentication.
    """

    AccessKeyId: str
    """
    The access key ID that identifies the temporary security credentials.
    """
    SecretAccessKey: str
    """
    The secret access key that can be used to sign requests.
    """
    SessionToken: str
    """
    The token that users must pass to the service API to use the temporary credentials.
    """
    Expiration: datetime
    """
    The date on which the current credentials expire.
    """


class STSAssumedRoleUser(Boto3Model):
    """
    The identifiers for the temporary security credentials that the operation returns.
    """

    AssumedRoleId: str
    """
    A unique identifier that contains the role ID and the role session name of the role
    that is being assumed.

    The role ID is generated by Amazon Web Services when the role is created.
    """
    Arn: str
    """
    The ARN of the temporary security credentials that are returned from the AssumeRole
    action.

    For more information about ARNs and how to use them in policies, see
    `IAM Identifiers <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html>`_
    in the *IAM User Guide*.
    """


class AssumeRole(ReadonlyPrimaryBoto3Model):
    """
    Contains the response to a successful AssumeRole request, including temporary Amazon
    Web Services credentials that can be used to make Amazon Web Services requests.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = AssumeRoleManager

    Credentials: STSCredentials | None = None
    """
    The temporary security credentials, which include an access key ID, a secret access
    key, and a security (or session) token.
    """
    AssumedRoleUser: STSAssumedRoleUser | None = None
    """
    The Amazon Resource Name (ARN) and the assumed role ID, which are identifiers that
    you can use to refer to the resulting temporary security credentials.

    For example, you can reference these credentials as a principal in a resource-based
    policy by using the ARN or assumed role ID. The ARN and ID include the ``RoleSessionName`` that you specified when you
    called ``AssumeRole``.
    """
    PackedPolicySize: int | None = None
    """
    A percentage value that indicates the packed size of the session policies and
    session tags combined passed in the request.

    The request fails if the packed size is greater than 100 percent, which means the
    policies and tags exceeded the allowed space.
    """
    SourceIdentity: str | None = None
    """
    The source identity specified by the principal that is calling the ``AssumeRole``
    operation.
    """
    Subject: str | None = None
    """
    The subject of the assumed role.
    """
    SubjectType: str | None = None
    """
    The type of the subject of the assumed role (SAML only).
    """
    Issuer: str | None = None
    """
    The issuer of the assumed role (SAML Only).
    """
    Provider: str | None = None
    """
    The provider of the assumed role (Web Identity Only).
    """
    Audience: str | None = None
    """
    The audience of the assumed role (SAML Only).
    """
    NameQualifier: str | None = None
    """
    The name qualifier of the assumed role (SAML Only).
    """
    SubjectFromWebIdentityToken: str | None = None
    """
    The subject from the web identity token (Web Identity Only).
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`Credentials.AccessKeyId` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.Credentials.AccessKeyId

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`AssumedRoleUser.Arn` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.AssumedRoleUser.Arn

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`Credentials.AccessKeyId` attribute.
        """
        return hash(self.Credentials.AccessKeyId)


class CallerIdentity(ReadonlyPrimaryBoto3Model):
    """
    Contains the response to a successful GetCallerIdentity request, including
    information about the entity making the request.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = CallerIdentityManager

    UserId: str | None = None
    """
    The unique identifier of the calling entity.

    The exact value depends on the type of entity that is making the call. The values
    returned are those listed in the **aws:userid** column in the
    `Principal table <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_variables.html#principaltable>`_
    found on the
    **Policy Variables** reference page in the *IAM User Guide*.
    """
    Account: str | None = None
    """
    The Amazon Web Services account ID number of the account that owns or contains the
    calling entity.
    """
    Arn: str | None = None
    """
    The Amazon Web Services ARN associated with the calling entity.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`Account` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.Account

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`Arn`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.Arn

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the :py:attr:`UserId`
        attribute.

        Returns:
            The name of the model instance.
        """
        return self.UserId

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`Account` attribute.
        """
        return hash(self.Account)


# =======================
# Request/Response Models
# =======================


class STSPolicyDescriptorType(Boto3Model):
    """
    A reference to the IAM managed policy that is passed as a session policy for a role
    session or a federated user session.
    """

    arn: str | None = None
    """
    The Amazon Resource Name (ARN) of the IAM managed policy to use as a session policy
    for the role.

    For more information
    about ARNs, see `Amazon Resource Names (ARNs) and Amazon Web Services Service
    Namespaces <https://docs.aws.amazon.com/general/latest/gr/aws-arns-and-namespaces.html>`_ in the *Amazon Web Services
    General Reference*.
    """


class STSProvidedContext(Boto3Model):
    """
    Contains information about the provided context.

    This includes the signed and encrypted trusted context assertion and the context
    provider ARN from which the trusted context assertion was generated.
    """

    ProviderArn: str | None = None
    """
    The context provider ARN from which the trusted context assertion was generated.
    """
    ContextAssertion: str | None = None
    """
    The signed and encrypted trusted context assertion generated by the context
    provider.

    The trusted context assertion is signed and encrypted by Amazon Web Services STS.
    """


class AssumeRoleWithSAMLResponse(Boto3Model):
    """
    Contains the response to a successful AssumeRoleWithSAML request, including
    temporary Amazon Web Services credentials that can be used to make Amazon Web
    Services requests.
    """

    Credentials: STSCredentials | None = None
    """
    The temporary security credentials, which include an access key ID, a secret access
    key, and a security (or session) token.
    """
    AssumedRoleUser: STSAssumedRoleUser | None = None
    """
    The identifiers for the temporary security credentials that the operation returns.
    """
    PackedPolicySize: int | None = None
    """
    A percentage value that indicates the packed size of the session policies and
    session tags combined passed in the request.

    The request fails if the packed size is greater than 100 percent, which means the
    policies and tags exceeded the allowed space.
    """
    Subject: str | None = None
    """
    The value of the ``NameID`` element in the ``Subject`` element of the SAML
    assertion.
    """
    SubjectType: str | None = None
    """
    The format of the name ID, as defined by the ``Format`` attribute in the ``NameID``
    element of the SAML assertion.

    Typical examples of the format are ``transient`` or ``persistent``.
    """
    Issuer: str | None = None
    """
    The value of the ``Issuer`` element of the SAML assertion.
    """
    Audience: str | None = None
    """
    The value of the ``Recipient`` attribute of the ``SubjectConfirmationData`` element
    of the SAML assertion.
    """
    NameQualifier: str | None = None
    """
A hash value based on the concatenation of the following:
    """
    SourceIdentity: str | None = None
    """
    The value in the ``SourceIdentity`` attribute in the SAML assertion.

    The source identity value persists across
    `chained role <https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles.html#iam-term-role-chaining>`_
    sessions.
    """


class AssumeRoleWithWebIdentityResponse(Boto3Model):
    """
    Contains the response to a successful AssumeRoleWithWebIdentity request, including
    temporary Amazon Web Services credentials that can be used to make Amazon Web
    Services requests.
    """

    Credentials: STSCredentials | None = None
    """
    The temporary security credentials, which include an access key ID, a secret access
    key, and a security token.
    """
    SubjectFromWebIdentityToken: str | None = None
    """
    The unique user identifier that is returned by the identity provider.

    This identifier is associated with the
    ``WebIdentityToken`` that was submitted with the ``AssumeRoleWithWebIdentity`` call. The identifier is typically unique
    to the user and the application that acquired the ``WebIdentityToken`` (pairwise identifier). For OpenID Connect ID
    tokens, this field contains the value returned by the identity provider as the token's ``sub`` (Subject) claim.
    """
    AssumedRoleUser: STSAssumedRoleUser | None = None
    """
    The Amazon Resource Name (ARN) and the assumed role ID, which are identifiers that
    you can use to refer to the resulting temporary security credentials.

    For example, you can reference these credentials as a principal in a resource-based
    policy by using the ARN or assumed role ID. The ARN and ID include the ``RoleSessionName`` that you specified when you
    called ``AssumeRole``.
    """
    PackedPolicySize: int | None = None
    """
    A percentage value that indicates the packed size of the session policies and
    session tags combined passed in the request.

    The request fails if the packed size is greater than 100 percent, which means the
    policies and tags exceeded the allowed space.
    """
    Provider: str | None = None
    """
    The issuing authority of the web identity token presented.

    For OpenID Connect ID tokens, this contains the value of the
    ``iss`` field. For OAuth 2.0 access tokens, this contains the value of the ``ProviderId`` parameter that was passed in
    the ``AssumeRoleWithWebIdentity`` request.
    """
    Audience: str | None = None
    """
    The intended audience (also known as client ID) of the web identity token.

    This is traditionally the client identifier issued to the application that requested
    the web identity token.
    """
    SourceIdentity: str | None = None
    """
    The value of the source identity that is returned in the JSON web token (JWT) from
    the identity provider.
    """
